import logging
import math
from itertools import chain
from multiprocessing import cpu_count
from typing import Optional

from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum
from joblib import delayed, Parallel

from fedot_ind.api.utils.input_data import init_input_data
from fedot_ind.core.metrics.metrics_implementation import *
from fedot_ind.core.models.quantile.stat_methods import stat_methods, stat_methods_global
from fedot_ind.core.operation.IndustrialCachableOperation import IndustrialCachableOperationImplementation
from fedot_ind.core.operation.transformation.data.hankel import HankelMatrix


class BaseExtractor(IndustrialCachableOperationImplementation):
    """
    Abstract class responsible for feature generator.
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.current_window = None
        self.stride = None
        self.n_processes = math.ceil(cpu_count() * 0.7) if cpu_count() > 1 else 1
        self.data_type = DataTypesEnum.table
        self.use_cache = params.get('use_cache', False)
        self.relevant_features = None
        self.logger = logging.getLogger(self.__class__.__name__)

        self.logging_params = {'jobs': self.n_processes}

    def fit(self, input_data: InputData):
        pass

    def extract_features(self, x, y) -> pd.DataFrame:
        """For those cases when you need to use feature extractor as a stangalone object
        """
        input_data = init_input_data(x, y)
        transformed_features = self.transform(input_data, use_cache=self.use_cache)
        try:
            return pd.DataFrame(transformed_features.predict, columns=self.relevant_features)
        except ValueError:
            return pd.DataFrame(transformed_features.predict)

    def _transform(self, input_data: InputData) -> np.array:
        """
        Method for feature generation for all series
        """
        features = input_data.features

        try:
            input_data_squeezed = np.squeeze(features, 3)
        except Exception:
            input_data_squeezed = np.squeeze(features)

        parallel = Parallel(n_jobs=self.n_processes, verbose=0, pre_dispatch="2*n_jobs")
        feature_matrix = parallel(delayed(self.generate_features_from_ts)(sample) for sample in input_data_squeezed)
        predict = self._clean_predict(np.array([ts.features for ts in feature_matrix]))
        self.relevant_features = feature_matrix[0].supplementary_data['feature_name']
        return predict

    @staticmethod
    def _clean_predict(predict: np.array):
        """Clean predict from nan, inf and reshape data for Fedot appropriate form
        """
        predict = np.where(np.isnan(predict), 0, predict)
        predict = np.where(np.isinf(predict), 0, predict)
        predict = predict.reshape(predict.shape[0], -1)
        return predict

    def generate_features_from_ts(self, ts_frame: np.array, window_length: int = None) -> np.array:
        """Method responsible for generation of features from time series.
        """
        pass

    @staticmethod
    def get_statistical_features(time_series: np.ndarray,
                                 add_global_features: bool = False) -> InputData:
        """
        Method for creating baseline quantile features for a given time series.

        Args:
            time_series: time series for which features are generated

        Returns:
            InputData: object with features

        """
        names = []
        vals = []
        time_series = time_series.flatten()

        if add_global_features:
            list_of_methods = [*stat_methods_global.items()]
        else:
            list_of_methods = [*stat_methods.items()]

        for method in list_of_methods:
            try:
                vals.append(method[1](time_series))
                names.append(method[0])
            except ValueError:
                continue

        stat_features = InputData(idx=np.arange(len(vals)),
                                  features=np.array(vals),
                                  target='no_target',
                                  task='no_task',
                                  data_type=DataTypesEnum.table,
                                  supplementary_data={'feature_name': names})
        return stat_features

    def apply_window_for_stat_feature(self, ts_data: np.array,
                                      feature_generator: callable,
                                      window_size: int = None) -> InputData:
        if window_size is None:
            # 10% of time series length by default
            window_size = round(ts_data.shape[0] / 10)
        else:
            window_size = round(ts_data.shape[0] * (window_size / 100))
        features = []
        names = []
        window_size = max(window_size, 5)
        self.stride = 3
        if self.stride > 1:
            trajectory_transformer = HankelMatrix(time_series=ts_data,
                                                  window_size=window_size,
                                                  strides=self.stride)
            subseq_set = trajectory_transformer.trajectory_matrix
        else:
            subseq_set = np.lib.stride_tricks.sliding_window_view(ts_data,
                                                                  ts_data.shape[0] - window_size)
        for i in range(0, subseq_set.shape[1]):
            slice_ts = subseq_set[:, i]
            stat_feature = feature_generator(slice_ts)
            features.append(stat_feature.features)
            names.append([x + f'_on_interval: {i + 1} - {i + 1 + window_size}'
                          for x in stat_feature.supplementary_data['feature_name']])
        window_stat_features = InputData(idx=np.arange(len(features)),
                                         features=features,
                                         target='no_target',
                                         task='no_task',
                                         data_type=DataTypesEnum.table,
                                         supplementary_data={'feature_name': names})
        return window_stat_features

    @staticmethod
    def _get_feature_matrix(extraction_func: callable,
                            ts: np.array) -> InputData:

        multi_ts_stat_features = [extraction_func(x) for x in ts]
        features = np.concatenate([component.features for component in multi_ts_stat_features], axis=0)

        for index, component in enumerate(multi_ts_stat_features):
            component.supplementary_data['feature_name'] = [f'{x} for component {index}'
                                                            for x in component.supplementary_data['feature_name']]
        names = list(
            chain(*[x.supplementary_data['feature_name'] for x in multi_ts_stat_features]))

        multi_ts_stat_features = InputData(idx=np.arange(len(features)),
                                           features=features,
                                           target='no_target',
                                           task='no_task',
                                           data_type=DataTypesEnum.table,
                                           supplementary_data={'feature_name': names})

        return multi_ts_stat_features



import numpy as np
from scipy.linalg import qr
from fedot_ind.core.operation.transformation.regularization.spectrum import singular_value_hard_threshold, \
    sv_to_explained_variance_ratio
import math


class RSVDDecomposition:
    def __init__(self,
                 rank: int = None
                 ):
        self.rank = rank

    def _init_random_params(self, tensor):
        # Percent of sampling columns. By default - 70%
        projection_rank = math.ceil(min(tensor.shape) / 1.5)
        # Polynom degree for power iteration procedure.
        self.poly_deg = 3
        # Create random matrix for projection/
        self.random_projection = np.random.randn(tensor.shape[1], projection_rank)

    def _compute_matrix_approximation(self, Ut, block, tensor, rank):
        Ut_ = Ut[:, :rank]
        tensor_approx = block @ Ut_
        reconstr_m = tensor_approx @ tensor_approx.T @ tensor
        return reconstr_m

    def _spectrum_regularization(self,
                                 spectrum: np.array,
                                 reg_type: str = 'hard_thresholding'):
        if reg_type == 'explained_dispersion':
            low_rank = sv_to_explained_variance_ratio(spectrum, 0)
        elif reg_type == 'hard_thresholding':
            low_rank = len(singular_value_hard_threshold(spectrum))
        return low_rank

    def _matrix_approx_regularization(self, low_rank, Ut, block, tensor):
        if low_rank == 1:
            return low_rank
        else:
            list_of_rank = list(range(1, low_rank + 1, 1))
            reconstr_matrix = [self._compute_matrix_approximation(Ut, block, tensor, rank) for rank in list_of_rank]
            fro_norms = [abs(np.linalg.norm(tensor - reconstr_m, 'fro')/np.linalg.norm(tensor)*100)
                         for reconstr_m in reconstr_matrix]
            deriviate_of_error = abs(np.diff(fro_norms))
            regularized_rank = len(deriviate_of_error[deriviate_of_error > 1]) + 1
        return regularized_rank

    def rsvd(self,
             tensor,
             approximation: bool = False,
             regularized_rank: int = None,
             reg_type: str = 'hard_thresholding') -> list:
        """Block Krylov subspace method for computing the SVD of a matrix with a low computational cost.

        Args:
            tensor (array (M, N) array_like):
            k (int): rank of the decomposition
            block_size (int): size of the block
            num_iter (int): number of iterations

        Returns:
            u, s, vt (array_like): decomposition

        Notes:
        :param reg_type:
        :param regularized_rank:
        :param approximation:

        """
        # Return classic svd decomposition with choosen type of spectrum thresholding
        if not approximation:
            # classic svd decomposition
            Ut, St, Vt = np.linalg.svd(tensor, full_matrices=False)
            # Compute low rank.
            low_rank = self._spectrum_regularization(St, reg_type=reg_type)
            if regularized_rank is not None:
                low_rank = regularized_rank
            # Return first n eigen components.
            U_, S_, V_ = Ut[:, :low_rank], St[:low_rank], Vt[:low_rank, :]
            return [U_, S_, V_]
        else:
            # First step. Initialize random matrix params.
            self._init_random_params(tensor)
            # Second step. Transform initial matrix to Gram. matrix
            AAT = tensor @ tensor.T
            # Third step. Power iteration procedure.First we raise the Gram matrix to the chosen degree.
            # This step is necessary in order to obtain a more "pronounced" spectrum (in which the eigenvalues
            # are well separated from each other). The important point is that the exponentiation procedure only changes
            # the eigenvalues but does not change the eigenvectors. Next, the resulting matrix is multiplied with the
            # original matrix ("overweighing" the column space) and then multiplied with a random matrix
            # in order to reduce the dimension and facilitate the procedure for "large" matrices.
            sampled_tensor = np.linalg.matrix_power(AAT, self.poly_deg) @ tensor @ self.random_projection
            # Fourth step. Orthogonalization of the resulting "sampled" matrix creates for us a basis of eigenvectors.
            sampled_tensor_orto, _ = qr(sampled_tensor, mode='economic')
            # Fifth step. Project initial Gramm matrix on new basis obtained from "sampled matrix".
            M = sampled_tensor_orto.T @ AAT @ sampled_tensor_orto
            # Six step. Classical svd decomposition with choosen type of spectrum thresholding
            Ut, St, Vt = np.linalg.svd(M, full_matrices=False)
            # Compute low rank.
            low_rank = self._spectrum_regularization(St, reg_type=reg_type)
            # Seven step. Compute matrix approximation and choose new low_rank
            if regularized_rank is None:
                regularized_rank = self._matrix_approx_regularization(low_rank, Ut, sampled_tensor_orto, tensor)
            # Eight step. Return matrix approximation.
            reconstr_tensor = self._compute_matrix_approximation(Ut, sampled_tensor_orto, tensor, regularized_rank)
            U_, S_, V_ = np.linalg.svd(reconstr_tensor, full_matrices=False)

            return [U_, S_, V_]



import math
import random
from typing import List, Tuple, Dict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


class TSTransformer:
    """
    Class for transformation single time series based on anomaly dictionary.

    Args:
        time_series: time series to split
        anomaly_dict: dictionary where keys are anomaly labels and values are anomaly index ranges.
        strategy: strategy for splitting time series. Available strategies: 'frequent' and `unique`.


    Attributes:
        selected_non_anomaly_intervals: list of non-anomaly intervals which were selected for splitting.

    Example:
        Create time series and anomaly dictionary::
            ts = np.random.rand(800)

        Or for multivariate time series::
            ts = [ts1, ts2, ts3]

        Then create anomaly dictionary::
            anomaly_d_uni = {'anomaly1': [[40, 50], [60, 80], [200, 220], [410, 420], [513, 524], [641, 645]],
                     'anomaly2': [[130, 170], [300, 320], [400, 410], [589, 620], [715, 720]],
                     'anomaly3': [[500, 530], [710, 740]],
                     'anomaly4': [[77, 90], [98, 112], [145, 158], [290, 322]]}

        Split time series into train and test parts::
            from fedot_ind.core.operation.transformation.splitter import TSSplitter
            splitter = TSSplitter(ts, anomaly_d)
            train, test = splitter.split(plot=True, binarize=False)

        In case of `unique` strategy, the splitting will be based on unique anomalies and hence
        the output of `split` method will be tuple of lists `unique_classes`, `unique_train`, `unique_test`
        where every element of every list is corresponding to unique anomaly. Important fact is that plotting
        function is now available for this case yet::
            unique_cls, unique_train, unique_test = splitter.split(strategy='unique', binarize=False)

    """

    def __init__(self, strategy: str = 'frequent'):

        self.strategy = strategy
        self.selected_non_anomaly_intervals = []
        self.split_methods = {'frequent': self._frequent_strategy,
                              'unique': self._unique_strategy}
        self.freq_length = None

    def __check_multivariate(self, time_series: np.ndarray):
        return isinstance(time_series, list) or (len(time_series.shape) > 1 and time_series.shape[1] > 1)

    def transform_for_fit(self, **kwargs):
        """ Splits data for train """
        method = self.split_methods.get(self.strategy, None)
        if method is None:
            raise ValueError(f'Unknown strategy {self.strategy} for splitting time series.')
        return method(**kwargs)

    def transform(self, series: np.array):
        """ Split test data on subsequences"""

        return self._transform_test(series)

    def _unique_strategy(self, series: np.array, anomaly_dict: Dict, plot: bool = False,
                         binarize: bool = False) -> tuple:
        """
        Split time series into train and test parts based on unique anomalies.

        Args:
            plot: if True, plot time series with anomaly intervals. Available only for univariate time series.
            binarize: if True, target will be binarized. Recommended for classification task if classes are imbalanced.

        Returns:
            tuple with train and test parts of time series ready for classification task with FedotIndustrial.

        """
        unique_classes, unique_trains = [], []
        for cls, list_of_inters in anomaly_dict.items():
            features, target = self.get_features_and_target(series=series,
                                                            classes=[cls],
                                                            transformed_intervals=[list_of_inters],
                                                            binarize=binarize)
            unique_trains.append((pd.DataFrame(features), np.array(target)))
            unique_classes.append(cls)

        return unique_classes, unique_trains

    def _frequent_strategy(self, series: np.array, anomaly_dict: Dict, plot: bool = False,
                           binarize: bool = False) -> tuple:
        """
        Method for splitting time series into train and test parts based on most frequent anomaly length.

        Args:
            plot: if True, plot time series with anomaly intervals.
            binarize: if True, target will be binarized. Recommended for classification task if classes are imbalanced.

        Returns:
            tuple with train and test parts of time series ready for classification task with FedotIndustrial.

        """
        classes = list(anomaly_dict.keys())
        intervals = list(anomaly_dict.values())
        self.freq_length = self._get_frequent_anomaly_length(intervals)
        transformed_intervals = self._transform_intervals(series, intervals)

        features, target = self.get_features_and_target(series=series,
                                                        classes=classes,
                                                        transformed_intervals=transformed_intervals,
                                                        binarize=binarize)

        if plot and not self.__check_multivariate(series):
            self.plot_classes_and_intervals(series=series,
                                            classes=classes,
                                            intervals=intervals,
                                            transformed_intervals=transformed_intervals)

        return features, target

    def get_features_and_target(self, series, classes, transformed_intervals, binarize) -> tuple:
        target, features = self._split_by_intervals(series, classes, transformed_intervals)
        non_anomaly_inters = self._get_non_anomaly_intervals(series, transformed_intervals)
        target, features = self.balance_with_non_anomaly(series, target, features, non_anomaly_inters)
        if binarize:
            target = self._binarize_target(target)
        return np.array(features), np.array(target)

    def _get_anomaly_intervals(self, anomaly_dict: Dict) -> Tuple[List[str], List[list]]:
        labels = list(anomaly_dict.keys())
        label_intervals = []
        for anomaly_label in labels:
            label_intervals.append(anomaly_dict[anomaly_label])
        return labels, label_intervals

    def _get_frequent_anomaly_length(self, intervals: List[list]):
        flat_intervals = []
        for sublist in intervals:
            for element in sublist:
                flat_intervals.append(element)

        lengths = list(map(lambda x: x[1] - x[0], flat_intervals))
        return max(set(lengths), key=lengths.count)

    def _transform_intervals(self, series, intervals):
        # ts = self.time_series if not self.multivariate else self.time_series.T[0]
        new_intervals = []
        for class_inter in intervals:
            new_class_intervals = []
            for i in class_inter:
                current_len = i[1] - i[0]
                abs_diff = abs(current_len - self.freq_length)
                left_add = math.ceil(abs_diff / 2)
                right_add = math.floor(abs_diff / 2)
                # Calculate new borders

                # If current anomaly interval is less than frequent length,
                # we expand current interval to the size of frequent
                if current_len < self.freq_length:
                    left = i[0] - left_add
                    right = i[1] + right_add
                    # If left border is negative, shift right border to the right
                    if left < 0:
                        right += abs(left)
                        left = 0
                    # If right border is greater than time series length, shift left border to the left
                    if right > len(series):
                        left -= abs(right - len(series))
                        right = len(series)

                # If current anomaly interval is greater than frequent length,
                # we shrink current interval to the size of frequent
                elif current_len > self.freq_length:
                    for l in range(i[0], i[1], self.freq_length):
                        new_class_intervals.append([l, l + self.freq_length])
                else:
                    left = i[0]
                    right = i[1]

                    new_class_intervals.append([left, right])
            new_intervals.append(new_class_intervals)
        return new_intervals

    def _split_by_intervals(self, series: np.array, classes: list, transformed_intervals: list) -> Tuple[
        List[str], List[list]]:
        all_labels, all_ts = [], []

        for i, label in enumerate(classes):
            for inter in transformed_intervals[i]:
                all_labels.append(label)
                if self.__check_multivariate(series):
                    all_ts.append(series[inter[0]:inter[1], :])
                else:
                    all_ts.append(np.ravel(series[inter[0]:inter[1]]))
        return all_labels, all_ts

    def plot_classes_and_intervals(self, series, classes, intervals, transformed_intervals):
        fig, axes = plt.subplots(3, 1, figsize=(17, 7))
        fig.tight_layout()
        for ax in axes:
            ax.plot(series, color='black', linewidth=1, alpha=1)

        axes[0].set_title('Initial intervals')
        axes[1].set_title('Transformed intervals')
        axes[2].set_title('Non-anomaly samples')

        for i, label in enumerate(classes):
            for interval_ in transformed_intervals[i]:
                axes[1].axvspan(interval_[0], interval_[1], alpha=0.3, color='blue')
                axes[1].text(interval_[0], 0.5, label, fontsize=12, rotation=90)
            for interval in intervals[i]:
                axes[0].axvspan(interval[0], interval[1], alpha=0.3, color='red')
                axes[0].text(interval[0], 0.5, label, fontsize=12, rotation=90)

        if self.selected_non_anomaly_intervals is not None:
            for interval in self.selected_non_anomaly_intervals:
                axes[2].axvspan(interval[0], interval[1], alpha=0.3, color='green')
                axes[2].text(interval[0], 0.5, 'no_anomaly', fontsize=12, rotation=90)
        plt.show()

    def _binarize_target(self, target):
        new_target = []
        for label in target:
            if label == 'no_anomaly':
                new_target.append(0)
            else:
                new_target.append(1)
        return new_target

    def balance_with_non_anomaly(self, series, target, features, non_anomaly_intervals):
        number_of_anomalies = len(target)
        anomaly_len = len(features[0])
        non_anomaly_ts_list = []
        ts = series.copy()
        counter = 0
        taken_slots = pd.Series([0 for _ in range(len(ts))])
        # for non_anom in non_anomaly_intervals:
        #     taken_slots[non_anom[0]:non_anom[1]] = 0

        while len(non_anomaly_ts_list) != number_of_anomalies and counter != number_of_anomalies * 100:
            seed = np.random.randint(1000)
            random.seed(seed)
            random_inter = random.choice(non_anomaly_intervals)
            cropped_ts_len = random_inter[1] - random_inter[0]
            counter += 1
            # Exclude intervals that are too short
            if cropped_ts_len < anomaly_len:
                continue
            random_start_index = random.randint(random_inter[0], random_inter[0] + cropped_ts_len - anomaly_len)
            stop_index = random_start_index + anomaly_len

            # Check if this interval overlaps with another interval
            if taken_slots[random_start_index:stop_index].mean() > 0.1:
                continue
            else:
                taken_slots[random_start_index:stop_index] = 1

            if self.__check_multivariate(series):
                non_anomaly_ts = ts[random_start_index:stop_index, :]
            else:
                non_anomaly_ts = np.ravel(ts[random_start_index:stop_index])

            non_anomaly_ts_list.append(non_anomaly_ts)

            self.selected_non_anomaly_intervals.append([random_start_index, stop_index])

        if len(non_anomaly_ts_list) == 0:
            raise Exception('No non-anomaly intervals found')

        target.extend(['no_anomaly'] * len(non_anomaly_ts_list))
        features.extend(non_anomaly_ts_list)

        return target, features

    def _get_non_anomaly_intervals(self, series, anom_intervals: List[list]):
        flat_intervals_list = []
        for sublist in anom_intervals:
            for element in sublist:
                flat_intervals_list.append(element)

        if self.__check_multivariate(series):
            series = pd.Series(series[:, 0]).copy()
        else:
            series = pd.Series(np.ravel(series)).copy()

        for single_interval in flat_intervals_list:
            series[single_interval[0]:single_interval[1]] = np.nan

        non_nan_intervals = []
        for k, g in series.groupby((series.notnull() != series.shift().notnull()).cumsum()):
            if g.notnull().any():
                non_nan_intervals.append((g.index[0], g.index[-1]))

        return non_nan_intervals

    def _transform_test(self, series: np.array):
        transformed_data = []
        for i in range(0, series.shape[0], self.freq_length):
            if len(series.shape) == 1:
                series_part = series[i:i + self.freq_length]
                if len(series_part) != self.freq_length:
                    series_part = series[-self.freq_length:]
            else:
                series_part = series[i:i + self.freq_length, :].T
                if series_part.shape[1] != self.freq_length:
                    series_part = series[-self.freq_length:, :].T
            transformed_data.append(series_part)
        transformed_data = np.stack(transformed_data)
        return transformed_data


if __name__ == '__main__':
    uni_ts = np.random.rand(800)
    anomaly_d_uni = {'anomaly1': [[40, 50], [60, 80], [200, 220], [410, 420], [513, 524], [641, 645]],
                     'anomaly2': [[130, 170], [300, 320], [400, 410], [589, 620], [715, 720]],
                     'anomaly3': [[500, 530], [710, 740]],
                     'anomaly4': [[77, 90], [98, 112], [145, 158], [290, 322]]}

    ts1 = np.arange(0, 100)
    multi_ts = np.array([ts1, ts1 * 2, ts1 * 3]).T
    anomaly_d_multi = {'anomaly1': [[0, 5], [15, 20], [22, 24], [55, 63], [70, 90]],
                       'anomaly2': [[10, 12], [15, 16], [27, 31], [44, 50], [98, 100]],
                       'anomaly3': [[0, 3], [15, 18], [19, 24], [55, 60], [85, 90]]}

    splitter_multi = TSTransformer()
    train_multi, test_multi = splitter_multi.transform_for_fit(series=multi_ts,
                                                               anomaly_dict=anomaly_d_multi, plot=False, binarize=True)

    splitter_uni = TSTransformer()
    train_uni, test_uni = splitter_uni.transform_for_fit(series=uni_ts,
                                                         anomaly_dict=anomaly_d_uni, plot=True, binarize=True)

    unique_ts = np.random.rand(800)
    anomaly_unique = {
        'class1': [[0, 10], [20, 30], [50, 60], [70, 80], [100, 110], [120, 130], [160, 170], 200, 210, 310, 330, 350,
                   370, 410, 430, 460, 480, 500, 520, [540, 560], [590, 610], [630, 650], [680, 700], [720, 740],
                   [760, 780], [80, 100], [320, 340]],
        'class2': [[0, 20], [50, 70], [100, 120], [140, 160], [190, 210], [230, 250], [270, 290], [240, 250],
                   [270, 280], [330, 340], [360, 370], [400, 410], [440, 450], [480, 490], [520, 530], [570, 580],
                   [610, 620], [660, 670], [700, 710]]}

    # splitter_unique = TSTransformer(strategy='unique')
    # unique_cls, unique_train, unique_test = splitter_unique.transform_for_fit(series=unique_ts,
    #                                                                           anomaly_dict=anomaly_unique, plot=True,
    #                                                                           binarize=False)
    _ = 1



from typing import Optional, Tuple, TypeVar

import numpy as np
import pandas as pd
import tensorly as tl
from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from joblib import delayed, Parallel
from pymonad.either import Either
from pymonad.list import ListMonad
from scipy import stats
from scipy.spatial.distance import cdist
from tensorly.decomposition import parafac
from tqdm import tqdm

from fedot_ind.core.operation.decomposition.matrix_decomposition.power_iteration_decomposition import RSVDDecomposition
from fedot_ind.core.operation.transformation.basis.abstract_basis import BasisDecompositionImplementation
from fedot_ind.core.operation.transformation.data.hankel import HankelMatrix
from fedot_ind.core.operation.transformation.regularization.spectrum import reconstruct_basis, \
    singular_value_hard_threshold

class_type = TypeVar("T", bound="DataDrivenBasis")


class EigenBasisImplementation(BasisDecompositionImplementation):
    """Eigen basis decomposition implementation
        Example:
            ts1 = np.random.rand(200)
            ts2 = np.random.rand(200)
            ts = [ts1, ts2]
            bss = EigenBasisImplementation({'window_size': 30})
            basis_multi = bss._transform(ts)
            basis_1d = bss._transform(ts1)
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.window_size = params.get('window_size', 20)
        self.low_rank_approximation = params.get('low_rank_approximation', True)
        self.basis = None
        self.SV_threshold = None
        self.svd_estimator = RSVDDecomposition()
        self.logging_params.update({'WS': self.window_size})

    def _combine_components(self, predict):
        count = 0
        grouped_v = []
        for df in predict:
            tmp = pd.DataFrame(df)
            ff = cdist(metric='cosine', XA=tmp.values, XB=tmp.values)
            if ff[-1, -2] < 0.5:
                count += 1
            tmp.iloc[-2, :] = tmp.iloc[-2,] + tmp.iloc[-1, :]
            tmp.drop(tmp.tail(1).index, inplace=True)
            grouped_v.append(tmp.values)

        if count / len(predict) > 0.35:
            self.SV_threshold = grouped_v[0].shape[0]
            self.logging_params.update({'SV_thr': self.SV_threshold})
            return np.array(grouped_v)
        else:
            return predict

    def _transform(self, input_data: InputData) -> np.array:
        """Method for transforming all samples

        """
        if isinstance(input_data, InputData):
            features = np.array(ListMonad(*input_data.features.tolist()).value)
        else:
            features = np.array(ListMonad(*input_data.values.tolist()).value)
        features = np.array([series[~np.isnan(series)] for series in features])

        if self.SV_threshold is None:
            self.SV_threshold = self.get_threshold(data=features)
            self.logging_params.update({'SV_thr': self.SV_threshold})

        parallel = Parallel(n_jobs=self.n_processes, verbose=0, pre_dispatch="2*n_jobs")
        v = parallel(delayed(self._transform_one_sample)(sample) for sample in features)
        predict = np.array(v)
        return predict

    def get_threshold(self, data) -> int:
        svd_numbers = []
        with tqdm(total=len(data), desc='SVD estimation') as pbar:
            for signal in data:
                svd_numbers.append(self._transform_one_sample(signal, svd_flag=True))
                pbar.update(1)
        return stats.mode(svd_numbers).mode[0]

    def _transform_one_sample(self, series: np.array, svd_flag: bool = False):
        trajectory_transformer = HankelMatrix(time_series=series, window_size=self.window_size)
        data = trajectory_transformer.trajectory_matrix
        self.ts_length = trajectory_transformer.ts_length
        if svd_flag:
            return self.estimate_singular_values(data)
        return self._get_basis(data)

    def estimate_singular_values(self, data):
        svd = lambda x: ListMonad(self.svd_estimator.rsvd(tensor=x, approximation=self.low_rank_approximation))
        basis = Either.insert(data).then(svd).value[0]
        spectrum = [s_val for s_val in basis[1] if s_val > 0.001]
        # self.left_approx_sv, self.right_approx_sv = basis[0], basis[2]
        return len(spectrum)

    def _get_1d_basis(self, data):
        data_driven_basis = lambda Monoid: ListMonad(reconstruct_basis(Monoid[0],
                                                                       Monoid[1],
                                                                       Monoid[2],
                                                                       ts_length=self.ts_length))
        threshold = lambda Monoid: ListMonad([Monoid[0],
                                              Monoid[1][:self.SV_threshold],
                                              Monoid[2]])
        svd = lambda x: ListMonad(self.svd_estimator.rsvd(tensor=x,
                                                          approximation=self.low_rank_approximation,
                                                          regularized_rank=self.SV_threshold))
        basis = Either.insert(data).then(svd).then(threshold).then(data_driven_basis).value[0]
        return np.swapaxes(basis, 1, 0)

    def _get_multidim_basis(self, data):
        rank = round(data[0].shape[0] / 10)
        beta = data[0].shape[0] / data[0].shape[1]

        tensor_decomposition = lambda x: ListMonad(parafac(tl.tensor(x), rank=rank).factors)
        multi_threshold = lambda x: singular_value_hard_threshold(singular_values=x,
                                                                  beta=beta,
                                                                  threshold=None)

        threshold = lambda Monoid: ListMonad([Monoid[1],
                                              list(map(multi_threshold, Monoid[0])),
                                              Monoid[2].T]) if self.n_components is None else ListMonad([Monoid[1][
                                                                                                         :,
                                                                                                         :self.n_components],
                                                                                                         Monoid[0][
                                                                                                         :,
                                                                                                         :self.n_components],
                                                                                                         Monoid[2][
                                                                                                         :,
                                                                                                         :self.n_components].T])
        data_driven_basis = lambda Monoid: ListMonad(reconstruct_basis(Monoid[0],
                                                                       Monoid[1],
                                                                       Monoid[2],
                                                                       ts_length=self.ts_length))

        basis = np.array(
            Either.insert(data).then(tensor_decomposition).then(threshold).then(data_driven_basis).value[0])

        basis = basis.reshape(basis.shape[1], -1)

        return basis

    def evaluate_derivative(self: class_type,
                            coefs: np.array,
                            order: int = 1) -> Tuple[class_type, np.array]:
        basis = type(self)(
            domain_range=self.domain_range,
            n_basis=self.n_basis - order,
        )
        derivative_coefs = np.array([np.polyder(x[::-1], order)[::-1] for x in coefs])

        return basis, derivative_coefs



import logging
from pathlib import Path
from typing import Tuple, Union

import numpy as np
import pandas as pd
from fedot.api.main import Fedot
from fedot.core.pipelines.pipeline import Pipeline

from fedot_ind.api.utils.configurator import Configurator
from fedot_ind.api.utils.reporter import ReporterTSC
from fedot_ind.core.architecture.experiment.computer_vision import CV_TASKS
from fedot_ind.core.architecture.settings.task_factory import TaskEnum
from fedot_ind.api.utils.path_lib import default_path_to_save_results
from fedot_ind.core.operation.transformation.splitter import TSTransformer
from fedot_ind.tools.synthetic.anomaly_generator import AnomalyGenerator
from fedot_ind.tools.synthetic.ts_generator import TimeSeriesGenerator


class FedotIndustrial(Fedot):
    """This class is used to run Fedot in industrial mode as FedotIndustrial.

    Args:
        input_config: dictionary with the parameters of the experiment.
        output_folder: path to the folder where the results will be saved.

    Example:
        First, configure experiment and instantiate FedotIndustrial class::
            from fedot_ind.api.main import FedotIndustrial
            from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader


            industrial = FedotIndustrial(task='ts_classification',
                                         dataset='ItalyPowerDemand',
                                         strategy='topological',
                                         use_cache=False,
                                         timeout=15,
                                         n_jobs=2,
                                         logging_level=20)

        Next, download data from UCR archive::

            train_data, test_data = DataLoader(dataset_name='ItalyPowerDemand').load_data()

        Finally, fit the model and get predictions::

            model = industrial.fit(train_features=train_data[0], train_target=train_data[1])
            labels = industrial.predict(test_features=test_data[0])
            probs = industrial.predict_proba(test_features=test_data[0])
            metric = industrial.get_metrics(target=test_data[1], metric_names=['f1', 'roc_auc'])

    """

    def __init__(self, **kwargs):
        kwargs.setdefault('output_folder', default_path_to_save_results())
        Path(kwargs.get('output_folder', default_path_to_save_results())).mkdir(parents=True, exist_ok=True)
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(name)s - %(message)s',
            handlers=[
                logging.FileHandler(Path(kwargs.get('output_folder')) / 'log.log'),
                logging.StreamHandler()
            ]
        )
        super(Fedot, self).__init__()

        self.logger = logging.getLogger('FedotIndustrialAPI')

        # self.reporter = ReporterTSC()
        self.configurator = Configurator()

        self.config_dict = None

        self.__init_experiment_setup(**kwargs)
        self.solver = self.__init_solver()

    def __init_experiment_setup(self, **kwargs):
        self.logger.info('Initialising experiment setup')
        # self.reporter.path_to_save = kwargs.get('output_folder')
        if 'task' in kwargs.keys() and kwargs['task'] in CV_TASKS.keys():
            self.config_dict = kwargs
        else:
            self.config_dict = self.configurator.init_experiment_setup(**kwargs)

    def __init_solver(self):
        self.logger.info('Initialising solver')

        if self.config_dict['task'] == 'ts_classification':
            if self.config_dict['strategy'] == 'fedot_preset':
                solver = TaskEnum[self.config_dict['task']].value['fedot_preset']
            # elif self.config_dict['strategy'] is None:
            #     self.config_dict['strategy'] = 'InceptionTime'
            #     solver = TaskEnum[self.config_dict['task']].value['nn']
            else:
                solver = TaskEnum[self.config_dict['task']].value['default']
        elif self.config_dict['task'] == 'ts_forecasting':
            if self.config_dict['strategy'] == 'decomposition':
                solver = TaskEnum[self.config_dict['task']].value['fedot_preset']

        else:
            solver = TaskEnum[self.config_dict['task']].value[0]
        return solver(self.config_dict)

    def fit(self, **kwargs) -> Pipeline:
        """
        Method for training Industrial model.

        Args:
            train_features: raw train data
            train_target: target values
            **kwargs: additional parameters

        Returns:
            :class:`Pipeline` object.

        """

        fitted_pipeline = self.solver.fit(**kwargs)
        return fitted_pipeline

    def predict(self, **kwargs) -> np.ndarray:
        """
        Method to obtain prediction labels from trained Industrial model.

        Args:
            test_features: raw test data

        Returns:
            the array with prediction values

        """
        return self.solver.predict(**kwargs)

    def predict_proba(self, **kwargs) -> np.ndarray:
        """
        Method to obtain prediction probabilities from trained Industrial model.

        Args:
            test_features: raw test data

        Returns:
            the array with prediction probabilities

        """
        return self.solver.predict_proba(**kwargs)

    def get_metrics(self, **kwargs) -> dict:
        """
        Method to obtain Gets quality metrics

        Args:
            target: target values
            metric_names: list of metric names desired to be calculated
            **kwargs: additional parameters

        Returns:
            the dictionary with calculated metrics

        """
        return self.solver.get_metrics(**kwargs)

    def save_predict(self, predicted_data: Union[pd.DataFrame, np.ndarray], **kwargs) -> None:
        """
        Method to save prediction locally in csv format

        Args:
            predicted_data: predicted data. For TSC task it could be either labels or probabilities

        Returns:
            None

        """
        kind = kwargs.get('kind')
        self.solver.save_prediction(predicted_data, kind=kind)

    def save_metrics(self, **kwargs) -> None:
        """
        Method to save metrics locally in csv format

        Args:
            metrics: dictionary with calculated metrics

        Returns:
            None

        """
        self.solver.save_metrics(**kwargs)

    def load(self, path):
        """Loads saved Industrial model from disk

        Args:
            path (str): path to the model

        """
        raise NotImplementedError()

    def plot_prediction(self, **kwargs):
        """Plot prediction of the model"""
        raise NotImplementedError()

    def explain(self, **kwargs):
        raise NotImplementedError()

    def generate_ts(self, ts_config: dict) -> np.array:
        """
        Method to generate synthetic time series
        Args:
            ts_config: dict with config for synthetic ts_data.

        Returns:
            synthetic time series data.

        """
        ts_generator = TimeSeriesGenerator(ts_config)
        t_series = ts_generator.get_ts()
        return t_series

    def generate_anomaly_ts(self,
                            ts_data: Union[dict, np.array],
                            anomaly_config: dict,
                            plot: bool = False,
                            overlap: float = 0.1) -> Tuple[np.array, np.array, dict]:
        """
        Method to generate anomaly time series
        Args:
            ts_data: either np.ndarray or dict with config for synthetic ts_data.
            anomaly_config: dict with config for anomaly generation
            overlap: float, ``default=0.1``. Defines the maximum overlap between anomalies.
            plot: if True, plot initial and modified time series data with rectangle spans of anomalies.

        Returns:
            returns initial time series data, modified time series data and dict with anomaly intervals.

        Returns:

        """

        generator = AnomalyGenerator(config=anomaly_config)
        init_synth_ts, mod_synth_ts, synth_inters = generator.generate(time_series_data=ts_data,
                                                                       plot=plot,
                                                                       overlap=overlap)

        return init_synth_ts, mod_synth_ts, synth_inters

    def split_ts(self, time_series: np.array,
                 anomaly_dict: dict,
                 binarize: bool = False,
                 strategy: str = 'frequent',
                 plot: bool = True) -> Tuple[np.array, np.array]:

        splitter = TSTransformer(time_series=time_series,
                                 anomaly_dict=anomaly_dict,
                                 strategy=strategy)

        train_data, test_data = splitter.transform_for_fit(plot=plot,
                                                           binarize=binarize)

        return train_data, test_data



"""This module contains classes for wrapping data of various types
for passing it to the prediction method of computer vision models.
"""

import os
from typing import Tuple, Callable

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset

IMG_EXTENSIONS = (".jpg", ".jpeg", ".png", ".ppm", ".bmp", ".pgm", ".tif", ".tiff", ".webp")


class PredictionNumpyDataset(Dataset):
    """Class for prediction on numpy arrays.

    Args:
        images: Numpy matrix of images.
    """

    def __init__(
            self,
            images: np.ndarray,
    ) -> None:
        self.images = torch.from_numpy(images).float()

    def __getitem__(self, idx) -> Tuple[torch.Tensor, int]:
        """Returns a sample from a dataset.

        Args:
            idx: Index of sample.

        Returns:
            A tuple ``(image, id)``, where image is image tensor,
                and id is integer.
        """
        return self.images[idx], idx

    def __len__(self) -> int:
        """Return length of dataset"""
        return self.images.size()[0]


class PredictionFolderDataset(Dataset):
    """Class for prediction on images from folder.

    Args:
        image_folder: Path to image folder.
        transform: A function/transform that takes in an PIL image and returns a
            transformed version.
    """

    def __init__(
            self,
            image_folder: str,
            transform: Callable,
    ) -> None:
        self.root = image_folder
        self.images = []
        for address, dirs, files in os.walk(image_folder):
            for name in files:
                if name.lower().endswith(IMG_EXTENSIONS):
                    self.images.append(os.path.join(address, name))
        self.transform = transform

    def __getitem__(self, idx) -> Tuple[torch.Tensor, str]:
        """Returns a sample from a dataset.

        Args:
            idx: Index of sample.

        Returns:
            A tuple ``(image, id)``, where image is image tensor,
                and id is file name.
        """

        image = Image.open(os.path.join(self.root, self.images[idx])).convert('RGB')
        image = self.transform(image)
        return image, self.images[idx]

    def __len__(self) -> int:
        """Return length of dataset"""
        return len(self.images)



class WindowCutter:
    """
    Window cutter.
        input format: dict with "data" and "labels" fields
        output: the same dict but with additional windows_list and labels for it
    """

    def __init__(self, window_len, window_step):
        # super().__init__(name="Window Cutter", operation="window cutting")
        self.input_dict = None
        self.window_len = window_len
        self.window_step = window_step
        self.output_window_list = []

    def load_data(self, input_dict: dict) -> None:
        self.input_dict = input_dict

    def get_windows(self) -> list:
        return self.output_window_list

    def run(self) -> None:
        """
        Cut data to windows
        :return: none
        """
        self.output_window_list = self._cut_ts_to_windows(self.input_dict)

    def _cut_ts_to_windows(self, ts: dict) -> list:
        start_idx = 0
        end_idx = len(ts[list(ts.keys())[0]]) - self.window_len
        temp_windows_list = []
        for i in range(start_idx, end_idx, self.window_step):
            temp_window_dict = {}
            for key in ts.keys():
                temp_window = []
                for j in range(i, i + self.window_len):
                    temp_window.append(ts[key][j])
                temp_window_dict[key] = temp_window
        temp_windows_list.append(temp_window_dict)
        return temp_windows_list



import logging
from typing import List, Union
from typing import Optional

import numpy as np
import pandas as pd
from fedot.api.main import Fedot
from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.pipelines.node import PipelineNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_builder import PipelineBuilder
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum
from golem.core.tuning.sequential import SequentialTuner

from fedot_ind.api.utils.input_data import init_input_data
from fedot_ind.api.utils.path_lib import default_path_to_save_results
from fedot_ind.api.utils.saver_collections import ResultSaver
from fedot_ind.core.metrics.evaluation import PerformanceAnalyzer
from fedot_ind.core.repository.initializer_industrial_models import IndustrialModels

np.random.seed(0)


class TimeSeriesClassifierPreset:
    """Class responsible for interaction with Fedot classifier. It allows to use FEDOT optimization
    for hyperparameters tuning and pipeline building. Nodes of the pipeline are basis functions
    from the list of branch_nodes and quantile extractor.

    Attributes:
        branch_nodes: list of nodes to be used in the pipeline
        tuning_iters: number of iterations for tuning hyperparameters of preprocessing pipeline
        model_params: parameters of the FEDOT classification model
        dataset_name: name of the dataset to be used
        saver: object of ``ResultSaver`` class

    Notes: ``branch_nodes`` can be one or combination of the following: ``'data_driven_basis'``, ``'fourier_basis'``,
           ``'wavelet_basis'``.

    """

    def __init__(self, params: Optional[OperationParameters] = None):
        self.branch_nodes: list = params.get('branch_nodes', None)

        self.model_params = params.get('model_params')
        self.dataset_name = params.get('dataset')
        self.tuning_iters = params.get('tuning_iterations', 30)
        self.tuning_timeout = params.get('tuning_timeout', 15.0)
        self.output_folder = params.get('output_folder', default_path_to_save_results())

        self.saver = ResultSaver(dataset_name=self.dataset_name,
                                 generator_name='fedot_preset',
                                 output_dir=self.output_folder)
        self.logger = logging.getLogger('TimeSeriesClassifier_Preset')

        self.test_predict_hash = None
        self.test_data_preprocessed = None
        self.prediction_label = None
        self.predictor = None
        self.y_train = None
        self.train_features = None
        self.test_features = None
        self.input_test_data = None
        self.preprocessing_pipeline = self._build_pipeline()

        self.logger.info(f'TimeSeriesClassifierPreset initialised with [{self.branch_nodes}] nodes and '
                         f'[{self.tuning_iters}] tuning iterations and [{self.tuning_timeout}] timeout')

    def _init_input_data(self, X: pd.DataFrame, y: np.ndarray) -> InputData:
        return init_input_data(X, y)

    def _build_pipeline(self):
        """
        Method for building pipeline with nodes from ``branch_nodes`` list and quantile extractor.

        """
        if self.branch_nodes is None:
            self.branch_nodes = ['eigen_basis', 'fourier_basis', 'wavelet_basis']

        self.extractors = ['quantile_extractor'] * len(self.branch_nodes)

        pipeline_builder = PipelineBuilder()

        for index, (basis, extractor) in enumerate(zip(self.branch_nodes, self.extractors)):
            pipeline_builder.add_node(basis, branch_idx=index)
            pipeline_builder.add_node(extractor, branch_idx=index)
        # pipeline_builder.join_branches('mlp', params={'hidden_layer_sizes': (256, 128, 64, 32),
        #                                               'max_iter': 300,
        #                                               'activation': 'relu',
        #                                               'solver': 'adam', })
        pipeline_builder.join_branches('rf')

        return pipeline_builder.build()

    def _tune_pipeline(self, pipeline: Pipeline, train_data: InputData):
        """
        Method for tuning pipeline with simultaneous tuner.

        Args:
            pipeline: pipeline to be tuned
            train_data: InputData object with train data

        Returns:
            tuned pipeline
        """
        if train_data.num_classes > 2:
            metric = ClassificationMetricsEnum.f1
        else:
            metric = ClassificationMetricsEnum.ROCAUC

        pipeline_tuner = TunerBuilder(train_data.task) \
            .with_tuner(SequentialTuner) \
            .with_metric(metric) \
            .with_timeout(self.tuning_timeout) \
            .with_iterations(self.tuning_iters) \
            .build(train_data)

        pipeline = pipeline_tuner.tune(pipeline)
        return pipeline

    def fit(self, features,
            target: np.ndarray = None,
            **kwargs) -> object:
        """
        Method for fitting pipeline on train data. It also tunes pipeline and updates it with categorical features.

        Args:
            features: pandas DataFrame with features
            target: numpy array with target values

        Returns:
            fitted FEDOT model as object of ``Pipeline`` class

        """

        with IndustrialModels():
            self.train_data = self._init_input_data(features, target)

            self.preprocessing_pipeline = self._tune_pipeline(self.preprocessing_pipeline,
                                                              self.train_data)
            self.preprocessing_pipeline.fit(self.train_data)

            self.baseline_model = self.preprocessing_pipeline.nodes[0]
            self.preprocessing_pipeline.update_node(self.baseline_model, PipelineNode('cat_features'))
            self.baseline_model.nodes_from = []
            # rf_node.unfit()
            self.preprocessing_pipeline.fit(self.train_data)

            train_data_preprocessed = self.preprocessing_pipeline.root_node.predict(self.train_data)
            train_data_preprocessed.predict = np.squeeze(train_data_preprocessed.predict)

            train_data_preprocessed = InputData(idx=train_data_preprocessed.idx,
                                                features=train_data_preprocessed.predict,
                                                target=train_data_preprocessed.target,
                                                data_type=train_data_preprocessed.data_type,
                                                task=train_data_preprocessed.task)

            self.processed_train_features = train_data_preprocessed.features

        metric = 'roc_auc' if train_data_preprocessed.num_classes == 2 else 'f1'
        self.model_params.update({'metric': metric})
        self.predictor = Fedot(available_operations=['scaling',
                                                     'normalization',
                                                     #'fast_ica',
                                                     'xgboost',
                                                     'rfr',
                                                     'rf',
                                                     'logit',
                                                     'mlp',
                                                     'knn',
                                                     'lgbm',
                                                     'pca'
                                                     ]
                               , **self.model_params)

        self.predictor.fit(train_data_preprocessed)

        return self.predictor

    def predict(self, features: pd.DataFrame, target: np.array) -> dict:
        """
        Method for prediction on test data.

        Args:
            features: pandas DataFrame with features
            target: numpy array with target values

        """

        test_data = self._init_input_data(features, target)
        test_data_preprocessed = self.preprocessing_pipeline.root_node.predict(test_data)

        if test_data.features.shape[0] == 1:
            test_data_preprocessed.predict = np.squeeze(test_data_preprocessed.predict).reshape(1, -1)
        else:
            test_data_preprocessed.predict = np.squeeze(test_data_preprocessed.predict)
        self.test_data_preprocessed = InputData(idx=test_data_preprocessed.idx,
                                                features=test_data_preprocessed.predict,
                                                target=test_data_preprocessed.target,
                                                data_type=test_data_preprocessed.data_type,
                                                task=test_data_preprocessed.task)

        self.prediction_label_baseline = self.baseline_model.predict(self.test_data_preprocessed).predict
        self.prediction_label = self.predictor.predict(self.test_data_preprocessed)

        return self.prediction_label

    def predict_proba(self, features, target) -> dict:
        test_data = self._init_input_data(features, target)
        test_data_preprocessed = self.preprocessing_pipeline.root_node.predict(test_data)
        self.test_data_preprocessed.predict = np.squeeze(test_data_preprocessed.predict)

        self.prediction_proba_baseline = self.baseline_model.predict(self.test_data_preprocessed, 'probs').predict
        self.prediction_proba = self.predictor.predict_proba(self.test_data_preprocessed)
        return self.prediction_proba

    def get_metrics(self, target: Union[np.ndarray, pd.Series], metric_names: Union[str, List[str]]) -> dict:
        """
        Method for calculating metrics on test data.

        Args:
            target: numpy array with target values
            metric_names: list of desirable metrics names

        Returns:
            dictionary with metrics values that looks like ``{metric_name: metric_value}``

        """
        analyzer = PerformanceAnalyzer()
        self.baseline_metrics = analyzer.calculate_metrics(target=np.ravel(target),
                                                           predicted_labels=self.prediction_label_baseline,
                                                           predicted_probs=self.prediction_proba_baseline,
                                                           target_metrics=metric_names)
        return analyzer.calculate_metrics(target=np.ravel(target),
                                          predicted_labels=self.prediction_label,
                                          predicted_probs=self.prediction_proba,
                                          target_metrics=metric_names)

    def save_prediction(self, predicted_data: np.ndarray, kind: str):
        self.saver.save(predicted_data, kind)

    def save_metrics(self, metrics: dict):
        self.saver.save(self.baseline_metrics, 'baseline_metrics')
        self.saver.save(metrics, 'metrics')



import numpy as np
from joblib import delayed, Parallel
from sklearn.decomposition import PCA
from sklearn.tree import DecisionTreeClassifier
from scipy.stats import mode

# An Implementation of the Rotation Forest Algorithm

# Potential Improvements:
# Multithreading on RotationForest.fit
# Clean up RotationTree.get_samples
from tqdm import tqdm


class RotationTree:
    """
    Base estimator for rotation forest

    Attributes:
        model: sklearn classifier used to make predictions
        n_features: k from forest, number of features for each PCA partition
        partitions: list of how the tree breaks up train data: [[fi, fk],[fi, fk]... etc.]
        rotation_matrices: the associated rotation matrix for each partition


    """

    @classmethod
    def from_model(cls, tree, n_features=3, sample_prop=0.5, bootstrap=True):
        # Allows for use of custom tree params/classifier models, so long as they adhere to sklearn style
        rt = RotationTree(n_features, sample_prop, bootstrap)
        rt.model = tree
        return rt

    def __init__(self, n_features=3, sample_prop=0.5, bootstrap=True):
        self.model = DecisionTreeClassifier(criterion="gini",
                                            splitter="best",
                                            max_depth=None,
                                            min_samples_split=2,
                                            min_samples_leaf=1,
                                            min_weight_fraction_leaf=0.,
                                            max_features=1.0,
                                            random_state=None,
                                            max_leaf_nodes=None,
                                            class_weight=None)
        self.n_features = n_features
        self.sample_prop = sample_prop
        self.bootstrap = bootstrap

        self.partitions = []
        self.rotation_matrices = []

    def fit(self, X, y):
        # Split up + Save features
        feature_partitions = self.partition_features(X)

        # Apply transform and save rotation matrices
        transformed_partitions = []
        for partition in feature_partitions:
            sampled_data = self.get_samples(partition, y)
            rotation_matrix = self.get_rotation_matrix(sampled_data)
            transformed_partitions.append(np.dot(partition, rotation_matrix))

        new_X = np.concatenate(transformed_partitions, axis=1)

        if self.bootstrap:
            xx, yy = self.boot_sample(new_X, y)
            self.model.fit(xx, yy)

        elif not self.bootstrap:
            self.model.fit(new_X, y)

        else:
            raise ValueError("Bootstrap not interpretable as Boolean")

    def partition_features(self, x):
        """Returns list of randomly partitioned features"""
        n_cols = x.shape[1]
        column_cases = [i for i in range(n_cols)]

        # Extract columns and randomly shuffle
        np.random.shuffle(column_cases)

        # Use stride slicing to handle cases where n_cols % n_features != 0
        case_output = [[m for m in column_cases[i::self.n_features]] for i in range(self.n_features)]
        feature_output = []
        for partition in case_output:
            feature_output.append(np.array([x[:, i] for i in partition]).T)

        self.partition_nums = case_output

        return feature_output

    def get_samples(self, featureset, y):
        # Returns Bootstrapped Samples from a subset of features

        # Concatenate features and y
        featureset = np.column_stack([featureset, y])
        n_rows = featureset.shape[0]

        # Choose classes to include
        unique_y = np.unique(y)
        include = []
        for cat in unique_y:
            i = np.random.uniform(size=1)
            if i > 0.5:
                include.append(cat)
        if len(include) == 0:
            include = unique_y

        # Remove non-included
        mask = np.isin(featureset[:, -1], include)
        rotation_seed = featureset[mask, :]

        # Sample from data
        cases = np.random.choice(int(rotation_seed.shape[0]), size=round(self.sample_prop * rotation_seed.shape[0]))
        rotation_seed = rotation_seed[cases, :]

        # Remove y-values
        sample_y = rotation_seed[:, -1]
        rotation_seed = np.delete(rotation_seed, -1, 1)

        return rotation_seed

    def get_rotation_matrix(self, samples):
        # Fits PCA and returns rotation matrix (matrix of eigen vectors)
        pca = PCA()
        pca.fit(samples)
        rotation_matrix = pca.components_
        self.rotation_matrices.append(rotation_matrix)

        return rotation_matrix

    def boot_sample(self, x, y):
        '''
        Sample from transformed data w/replacement, return X, y;
        Could be static/stand-alone but I'm unsure of best practice and this is readable
        '''
        newdata = np.concatenate((x, y[:, np.newaxis]), axis=1)
        cases = np.random.choice(newdata.shape[0], size=newdata.shape[0], replace=True)
        samples = newdata[cases,]

        return samples[:, :-1], samples[:, -1]

    def predict(self, X):
        partitions = [np.array([X[:, i] for i in partition]).T for partition in self.partition_nums]

        transformed_partitions = []
        for i, p in enumerate(partitions):
            transformed_partitions.append(np.dot(p, self.rotation_matrices[i]))

        new_X = np.concatenate(transformed_partitions, axis=1)
        return self.model.predict(new_X)


class RotationForest:
    """

    Algorithm per https://arxiv.org/abs/1809.06705
    Input: k, the number of trees, f, the number of features, p, the sample proportion
    1: Let F =< F1 . . . Fk > be the C4.5 trees in the forest.
    2: for i ← 1 to k do
        3: Randomly partition the original features into r subsets, each with f features (r =
        m/f), denoted < S1 . . . Sr >.
        4: Let Di be the train set for tree i, initialised to the original data, Di ← D.
        5: for j ← 1 to r do
            6: Select a non-empty subset of classes and extract only cases with those class labels.
            Each class has 0.5 probability of inclusion.
            7: Draw a proportion p of cases (without replacement) of those with the selected
            class value
            8: Perform a Principal Component Analysis (PCA) on the features in Sj on this
            subset of data
            9: Apply the PCA transform built on this subset to the features in Sj of the whole
            train set
            10: Replace the features Sj in Di with the PCA features.
        11: Build C4.5 Classifier Fi on transformed data Di.
    """

    def __init__(self, n_trees=1000, percent_features=10, sample_prop=0.5, bootstrap=False):
        self.bootstrap = bootstrap
        self.n_trees = n_trees
        self.n_features = None
        self.percent_features = percent_features
        self.sample_prop = sample_prop

        self.is_fit = False
        self.trees = []
        pass

    def fit(self, X, y, model=None):
        self.n_features = round(X.shape[1] * (self.percent_features / 100))
        X = np.asarray(X)
        if model:
            for i in range(self.n_trees):
                tree = RotationTree.from_model(model,
                                               n_features=self.n_features,
                                               sample_prop=self.sample_prop,
                                               bootstrap=self.bootstrap)
                tree.fit(X, y)
                self.trees.append(tree)
        else:

            with tqdm(total=self.n_trees, unit='tree', desc='RotF training') as pbar:
                for i in range(self.n_trees):
                    tree = RotationTree(n_features=self.n_features,
                                        sample_prop=self.sample_prop,
                                        bootstrap=self.bootstrap)
                    tree.fit(X, y)
                    self.trees.append(tree)
                    pbar.update(1)
            # def fit_tree(i):
            #     tree = RotationTree(n_features=self.n_features,
            #                         sample_prop=self.sample_prop,
            #                         bootstrap=self.bootstrap)
            #     tree.fit(X, y)
            #     return tree
            #
            # with tqdm(total=self.n_trees, unit='tree', desc='RotF training') as pbar:
            #     self.trees = Parallel(n_jobs=4)(
            #         delayed(fit_tree)(i) for i in range(self.n_trees)
            #     )
            #     pbar.update(self.n_trees)

    def predict(self, X):
        X = np.asarray(X)
        # Generate predictions array
        all = []
        for model in self.trees:
            all.append(model.predict(X))
        all = np.asarray(all)
        preds = mode(all)[0].flatten()

        return preds


# RF Testing
if __name__ == "__main__":
    # Somewhat hacky test to play around with sci-kit generated data sets;
    # Under the current settings, Rotation Forest will usually (but not exclusively) outperform Random Forest.
    # However, it only performs better in some subset of all possible classification problems.

    test_forest = True

    if test_forest:
        from sklearn.datasets import make_classification
        from sklearn.model_selection import train_test_split
        from sklearn.ensemble import RandomForestClassifier

        x, y = make_classification(n_features=10, n_samples=1000, n_classes=2, n_clusters_per_class=6, n_informative=8)
        X_train, X_test, y_train, y_test = train_test_split(x, y, test_size=0.3)

        rf = RotationForest(n_trees=100, n_features=3, bootstrap=True)
        rf.fit(X_train, y_train)

        predictions = rf.predict(X_test)

        import pandas as pd

        results = pd.DataFrame([predictions, y_test], index=['Predicted', 'Observed']).T

        results['iseq'] = results.Observed.eq(results.Predicted).astype(int)

        correct = (results.iseq.sum() / len(results))
        print(f"Rotation Forest %Correct: {correct * 100}%")

        rf2 = RandomForestClassifier(n_estimators=100)
        rf2.fit(X_train, y_train)
        preds = rf2.predict(X_test)

        results = pd.DataFrame([preds, y_test], index=['Predicted', 'Observed']).T

        results['iseq'] = results.Observed.eq(results.Predicted).astype(int)
        correct = (results.iseq.sum() / len(results))
        print(f"Random Forest %Correct: {correct * 100}%")

    else:
        testtree = RotationTree()
        from sklearn.datasets import make_classification
        from sklearn.model_selection import train_test_split

        x, y = make_classification(n_features=50, n_samples=1000)
        trx, tex, trainy, testy = train_test_split(x, y)
        testtree.fit(trx, trainy)
        preds = testtree.predict(tex)
        obs = testy

        import pandas as pd

        df = pd.DataFrame([preds, obs]).T
        df['iseq'] = df[0].eq(df[1]).astype(int)
        print(df.iseq.sum() / len(df))



import gc
import sys
from functools import partial
from typing import Optional

import numpy as np
import pandas as pd
from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum

from gtda.time_series import takens_embedding_optimal_parameters
from scipy import stats
from tqdm import tqdm

from fedot_ind.core.models.base_extractor import BaseExtractor
from fedot_ind.core.models.topological.topofeatures import AverageHoleLifetimeFeature, \
    AveragePersistenceLandscapeFeature, BettiNumbersSumFeature, HolesNumberFeature, MaxHoleLifeTimeFeature, \
    PersistenceDiagramsExtractor, PersistenceEntropyFeature, RadiusAtMaxBNFeature, RelevantHolesNumber, \
    SimultaneousAliveHolesFeature, SumHoleLifetimeFeature, TopologicalFeaturesExtractor
from fedot_ind.core.operation.transformation.data.point_cloud import TopologicalTransformation

sys.setrecursionlimit(1000000000)

PERSISTENCE_DIAGRAM_FEATURES = {'HolesNumberFeature': HolesNumberFeature(),
                                'MaxHoleLifeTimeFeature': MaxHoleLifeTimeFeature(),
                                'RelevantHolesNumber': RelevantHolesNumber(),
                                'AverageHoleLifetimeFeature': AverageHoleLifetimeFeature(),
                                'SumHoleLifetimeFeature': SumHoleLifetimeFeature(),
                                'PersistenceEntropyFeature': PersistenceEntropyFeature(),
                                'SimultaneousAliveHolesFeature': SimultaneousAliveHolesFeature(),
                                'AveragePersistenceLandscapeFeature': AveragePersistenceLandscapeFeature(),
                                'BettiNumbersSumFeature': BettiNumbersSumFeature(),
                                'RadiusAtMaxBNFeature': RadiusAtMaxBNFeature()}

PERSISTENCE_DIAGRAM_EXTRACTOR = PersistenceDiagramsExtractor(takens_embedding_dim=1,
                                                             takens_embedding_delay=2,
                                                             homology_dimensions=(0, 1),
                                                             parallel=False)


class TopologicalExtractor(BaseExtractor):
    """Class for extracting topological features from time series data.

    Args:
        params: parameters for operation

    Example:
        To use this operation you can create pipeline as follows::
            from fedot.core.pipelines.pipeline_builder import PipelineBuilder
            from fedot_ind.api.utils.input_data import init_input_data
            from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
            from fedot_ind.core.repository.initializer_industrial_models import IndustrialModels

            train_data, test_data = DataLoader(dataset_name='Ham').load_data()
            with IndustrialModels():
                pipeline = PipelineBuilder().add_node('eigen_basis').add_node('topological_extractor').add_node(
                    'rf').build()
                input_data = init_input_data(train_data[0], train_data[1])
                pipeline.fit(input_data)
                features = pipeline.predict(input_data)
                print(features)
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.window_size = params.get('window_size', 10)
        self.stride = params.get('window_size', 1)
        self.feature_extractor = TopologicalFeaturesExtractor(
            persistence_diagram_extractor=PERSISTENCE_DIAGRAM_EXTRACTOR,
            persistence_diagram_features=PERSISTENCE_DIAGRAM_FEATURES)
        self.data_transformer = None

    def __evaluate_persistence_params(self, ts_data: np.array) -> InputData:
        if self.feature_extractor is None:
            te_dimension, te_time_delay = self.get_embedding_params_from_batch(ts_data=ts_data)

            persistence_diagram_extractor = PersistenceDiagramsExtractor(takens_embedding_dim=te_dimension,
                                                                         takens_embedding_delay=te_time_delay,
                                                                         homology_dimensions=(0, 1, 2),
                                                                         parallel=True)

            self.feature_extractor = TopologicalFeaturesExtractor(
                persistence_diagram_extractor=persistence_diagram_extractor,
                persistence_diagram_features=PERSISTENCE_DIAGRAM_FEATURES)

    def _generate_features_from_ts(self, ts_data: np.array,
                                   persistence_params: dict) -> InputData:
        if self.data_transformer is None:
            self.data_transformer = TopologicalTransformation(
                persistence_params=persistence_params,
                window_length=round(ts_data.shape[0] * 0.01 * self.window_size))

        point_cloud = self.data_transformer.time_series_to_point_cloud(input_data=ts_data)
        topological_features = self.feature_extractor.transform(point_cloud)
        topological_features = InputData(idx=np.arange(len(topological_features.values)),
                                         features=topological_features.values,
                                         target='no_target',
                                         task='no_task',
                                         data_type=DataTypesEnum.table,
                                         supplementary_data={'feature_name': topological_features.columns})
        return topological_features

    def generate_topological_features(self, ts: np.array,
                                      persistence_params: dict = None) -> InputData:

        if persistence_params is not None:
            self.__evaluate_persistence_params(ts)

        if len(ts.shape) == 1:
            aggregation_df = self._generate_features_from_ts(ts, persistence_params)
        else:
            aggregation_df = self._get_feature_matrix(partial(self._generate_features_from_ts,
                                                              persistence_params=persistence_params), ts)

        return aggregation_df

    def generate_features_from_ts(self, ts_data: np.array, dataset_name: str = None):
        return self.generate_topological_features(ts=ts_data)

    def get_embedding_params_from_batch(self, ts_data: pd.DataFrame, method: str = 'mean') -> tuple:
        """Method for getting optimal Takens embedding parameters.
        Args:
            ts_data: dataframe with time series data
            method: method for getting optimal parameters
        Returns:
            Optimal Takens embedding parameters
        """
        methods = {'mode': self._mode,
                   'mean': np.mean,
                   'median': np.median}

        dim_list, delay_list = list(), list()

        for _ in tqdm(range(len(ts_data)),
                      initial=0,
                      desc='Time series processed: ',
                      unit='ts', colour='black'):
            ts_data = pd.DataFrame(ts_data)
            single_time_series = ts_data.sample(1, replace=False, axis=0).squeeze()
            delay, dim = takens_embedding_optimal_parameters(X=single_time_series,
                                                             max_time_delay=1,
                                                             max_dimension=5,
                                                             n_jobs=-1)
            delay_list.append(delay)
            dim_list.append(dim)

        dimension = int(methods[method](dim_list))
        delay = int(methods[method](delay_list))

        return dimension, delay

    @staticmethod
    def _mode(arr: list) -> int:
        return int(stats.mode(arr)[0][0])



from __future__ import division, print_function

import numpy as np


class RecurrenceFeatureExtractor:
    def __init__(self, recurrence_matrix: np.ndarray = None):
        self.recurrence_matrix = recurrence_matrix

    def quantification_analysis(self, MDL: int = 3, MVL: int = 3, MWVL: int = 2):

        n_vectors = self.recurrence_matrix.shape[0]
        recurrence_rate = float(np.sum(self.recurrence_matrix)) / np.power(n_vectors, 2)

        diagonal_frequency_dist = self.calculate_diagonal_frequency(number_of_vectors=n_vectors)
        vertical_frequency_dist = self.calculate_vertical_frequency(number_of_vectors=n_vectors, not_white=1)
        white_vertical_frequency_dist = self.calculate_vertical_frequency(number_of_vectors=n_vectors,
                                                                          not_white=0)

        determinism = self.laminarity_or_determinism(MDL, n_vectors, diagonal_frequency_dist, lam=False)
        laminarity = self.laminarity_or_determinism(MVL, n_vectors, vertical_frequency_dist, lam=True)

        average_diagonal_line_length = self.average_line_length(MDL, n_vectors, diagonal_frequency_dist)
        average_vertical_line_length = self.average_line_length(MVL, n_vectors, vertical_frequency_dist)
        average_white_vertical_line_length = self.average_line_length(MWVL, n_vectors, white_vertical_frequency_dist)

        longest_diagonal_line_length = self.longest_line_length(diagonal_frequency_dist, n_vectors, diag=True)
        longest_vertical_line_length = self.longest_line_length(vertical_frequency_dist, n_vectors, diag=False)
        longest_white_vertical_line_length = self.longest_line_length(white_vertical_frequency_dist,
                                                                      n_vectors, diag=False)

        entropy_diagonal_lines = self.entropy_lines(MDL, n_vectors, diagonal_frequency_dist, diag=True)
        entropy_vertical_lines = self.entropy_lines(MVL, n_vectors, vertical_frequency_dist, diag=False)
        entropy_white_vertical_lines = self.entropy_lines(MWVL, n_vectors,
                                                          white_vertical_frequency_dist, diag=False)

        return {'RR': recurrence_rate, 'DET': determinism, 'ADLL': average_diagonal_line_length,
                'LDLL': longest_diagonal_line_length, 'DIV': 1. / longest_diagonal_line_length,
                'EDL': entropy_diagonal_lines, 'LAM': laminarity, 'AVLL': average_vertical_line_length,
                'LVLL': longest_vertical_line_length, 'EVL': entropy_vertical_lines,
                'AWLL': average_white_vertical_line_length, 'LWLL': longest_white_vertical_line_length,
                'EWLL': entropy_white_vertical_lines, 'RDRR': determinism / recurrence_rate,
                'RLD': laminarity / determinism}

    def calculate_vertical_frequency(self, number_of_vectors, not_white: int):
        vertical_frequency_distribution = np.zeros(number_of_vectors + 1)
        for i in range(number_of_vectors):
            vertical_line_length = 0
            for j in range(number_of_vectors):
                if self.recurrence_matrix[i, j] == not_white:
                    vertical_line_length += 1
                    if j == (number_of_vectors - 1):
                        vertical_frequency_distribution[vertical_line_length] += 1.0
                else:
                    if vertical_line_length != 0:
                        vertical_frequency_distribution[vertical_line_length] += 1.0
                        vertical_line_length = 0
        return vertical_frequency_distribution

    def calculate_diagonal_frequency(self, number_of_vectors):
        diagonal_frequency_distribution = np.zeros(number_of_vectors + 1)
        for i in range(number_of_vectors - 1, -1, -1):
            diagonal_line_length = 0
            for j in range(0, number_of_vectors - i):
                if self.recurrence_matrix[i + j, j] == 1:
                    diagonal_line_length += 1
                    if j == (number_of_vectors - i - 1):
                        diagonal_frequency_distribution[diagonal_line_length] += 1.0
                else:
                    if diagonal_line_length != 0:
                        diagonal_frequency_distribution[diagonal_line_length] += 1.0
                        diagonal_line_length = 0

        for k in range(1, number_of_vectors):
            diagonal_line_length = 0
            for i in range(number_of_vectors - k):
                j = i + k
                if self.recurrence_matrix[i, j] == 1:
                    diagonal_line_length += 1
                    if j == (number_of_vectors - 1):
                        diagonal_frequency_distribution[diagonal_line_length] += 1.0
                else:
                    if diagonal_line_length != 0:
                        diagonal_frequency_distribution[diagonal_line_length] += 1.0
                        diagonal_line_length = 0
        return diagonal_frequency_distribution

    def entropy_lines(self, factor, number_of_vectors, distribution, diag: bool):
        if diag:
            sum_frequency_distribution = float(np.sum(distribution[factor:-1]))
        else:
            number_of_vectors = number_of_vectors + 1
            sum_frequency_distribution = float(np.sum(distribution[factor:]))

        entropy_lines = 0
        for i in range(factor, number_of_vectors):
            if distribution[i] != 0:
                entropy_lines += (distribution[i] / sum_frequency_distribution) * \
                                 np.log(distribution[i] / sum_frequency_distribution)
        return -entropy_lines

    def laminarity_or_determinism(self, factor, number_of_vectors, distribution, lam: bool):
        if lam:
            number_of_vectors = number_of_vectors + 1
        numerator = np.sum([i * distribution[i] for i in range(factor, number_of_vectors)])
        denominator = np.sum([i * distribution[i] for i in range(1, number_of_vectors)])
        return numerator / denominator

    def longest_line_length(self, frequency_distribution, number_of_vectors, diag: bool):
        longest_line_length = 1
        for i in range(number_of_vectors, 0, -1):
            if frequency_distribution[i] != 0:
                return i
        return longest_line_length

    def average_line_length(self, factor, number_of_vectors, distribution):
        numerator = np.sum([i * distribution[i] for i in range(factor, number_of_vectors + 1)])
        denominator = np.sum([distribution[i] for i in range(factor, number_of_vectors + 1)])
        return numerator / denominator



from typing import Optional
import torch
from torch.nn import Conv2d, Parameter
from torch.nn.functional import conv2d

from fedot_ind.core.architecture.abstraction.сheckers import parameter_value_check


class DecomposedConv2d(Conv2d):
    """Extends the Conv2d layer by implementing the singular value decomposition of
    the weight matrix.

    Args:
        base_conv:  The convolutional layer whose parameters will be copied
        decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
            If ``None`` create layers without decomposition.
        forward_mode: ``'one_layer'``, ``'two_layers'`` or ``'three_layers'`` forward pass calculation method.
    """

    def __init__(
        self,
        base_conv: Conv2d,
        decomposing_mode: Optional[str] = 'channel',
        forward_mode: str = 'one_layer',
        device=None,
        dtype=None,
    ) -> None:

        parameter_value_check('forward_mode', forward_mode, {'one_layer', 'two_layers', 'three_layers'})

        if forward_mode != 'one_layer':
            assert base_conv.padding_mode == 'zeros', \
                "only 'zeros' padding mode is supported for '{forward_mode}' forward mode."
            assert base_conv.groups == 1, f"only 1 group is supported for '{forward_mode}' forward mode."

        super().__init__(
            base_conv.in_channels,
            base_conv.out_channels,
            base_conv.kernel_size,
            base_conv.stride,
            base_conv.padding,
            base_conv.dilation,
            base_conv.groups,
            (base_conv.bias is not None),
            base_conv.padding_mode,
            device,
            dtype,
        )
        self.load_state_dict(base_conv.state_dict())
        self.forward_mode = forward_mode
        if decomposing_mode is not None:
            self.decompose(decomposing_mode)
        else:
            self.U = None
            self.S = None
            self.Vh = None
            self.decomposing = None

    def __set_decomposing_params(self, decomposing_mode):
        n, c, w, h = self.weight.size()
        decomposing_modes = {
            'channel': {
                'type': 'channel',
                'permute': (0, 1, 2, 3),
                'decompose_shape': (n, c * w * h),
                'compose_shape': (n, c, w, h),
                'U shape': (n, 1, 1, -1),
                'U': {
                    'stride': 1,
                    'padding': 0,
                    'dilation': 1,
                },
                'Vh shape': (-1, c, w, h),
                'Vh': {
                    'stride': self.stride,
                    'padding': self.padding,
                    'dilation': self.dilation,
                }
            },
            'spatial': {
                'type': 'spatial',
                'permute': (0, 2, 1, 3),
                'decompose_shape': (n * w, c * h),
                'compose_shape': (n, w, c, h),
                'U shape': (n, w, 1, -1),
                'U': {
                    'stride': (self.stride[0], 1),
                    'padding': (self.padding[0], 0),
                    'dilation': (self.dilation[0], 1),
                },
                'Vh shape': (-1, c, 1, h),
                'Vh': {
                    'stride': (1, self.stride[1]),
                    'padding': (0, self.padding[1]),
                    'dilation': (1, self.dilation[1]),
                }
            },
        }
        parameter_value_check('decomposing_mode', decomposing_mode, set(decomposing_modes.keys()))
        self.decomposing = decomposing_modes[decomposing_mode]

    def decompose(self, decomposing_mode: str) -> None:
        """Decomposes the weight matrix in singular value decomposition.
        Replaces the weights with U, S, Vh matrices such that weights = U * S * Vh.
        Args:
            decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
        Raises:
            ValueError: If ``decomposing_mode`` not in valid values.
        """
        self.__set_decomposing_params(decomposing_mode=decomposing_mode)
        W = self.weight.permute(self.decomposing['permute']).reshape(self.decomposing['decompose_shape'])
        U, S, Vh = torch.linalg.svd(W, full_matrices=False)
        self.U = Parameter(U)
        self.S = Parameter(S)
        self.Vh = Parameter(Vh)
        self.register_parameter('weight', None)

    def compose(self) -> None:
        """Compose the weight matrix from singular value decomposition.
        Replaces U, S, Vh matrices with weights such that weights = U * S * Vh.
        """
        W = self.U @ torch.diag(self.S) @ self.Vh
        self.weight = Parameter(W.reshape(self.decomposing['compose_shape']).permute(self.decomposing['permute']))
        self.register_parameter('U', None)
        self.register_parameter('S', None)
        self.register_parameter('Vh', None)
        self.decomposing = None

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.decomposing is not None:
            if self.forward_mode == 'one_layer':
                return self._one_layer_forward(input)
            if self.forward_mode == 'two_layers':
                return self._two_layers_forward(input)
            if self.forward_mode == 'three_layers':
                return self._three_layers_forward(input)
        else:
            return self._conv_forward(input, self.weight, self.bias)

    def _one_layer_forward(self, input: torch.Tensor) -> torch.Tensor:
        W = self.U @ torch.diag(self.S) @ self.Vh
        W = W.reshape(self.decomposing['compose_shape']).permute(self.decomposing['permute'])
        return self._conv_forward(input, W, self.bias)

    def _two_layers_forward(self, input: torch.Tensor) -> torch.Tensor:
        SVh = (torch.diag(self.S) @ self.Vh).view(self.decomposing['Vh shape'])
        U = self.U.reshape(self.decomposing['U shape']).permute(0, 3, 1, 2)
        x = conv2d(input=input, weight=SVh, groups=self.groups, **self.decomposing['Vh'])
        return conv2d(input=x, weight=U, bias=self.bias, **self.decomposing['U'])

    def _three_layers_forward(self, input: torch.Tensor) -> torch.Tensor:
        S = torch.diag(self.S).view([len(self.S), len(self.S), 1, 1])
        Vh = self.Vh.view(self.decomposing['Vh shape'])
        U = self.U.view(self.decomposing['U shape']).permute(0, 3, 1, 2)
        x = conv2d(input=input, weight=Vh, groups=self.groups, **self.decomposing['Vh'])
        x = conv2d(input=x, weight=S, padding=0)
        return conv2d(input=x, weight=U, bias=self.bias, **self.decomposing['U'])

    def set_U_S_Vh(self, u: torch.Tensor, s: torch.Tensor, vh: torch.Tensor) -> None:
        """Update U, S, Vh matrices.
        Raises:
            Assertion Error: If ``self.decomposing`` is False.
        """
        assert self.decomposing is not None, "for setting U, S and Vh, the model must be decomposed"
        self.U = Parameter(u)
        self.S = Parameter(s)
        self.Vh = Parameter(vh)



import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from fedot_ind.tools.synthetic.ts_generator import TimeSeriesGenerator


class TimeSeriesDatasetsGenerator:
    def __init__(self, num_samples: int = 80, max_ts_len: int = 50, n_classes: int = 3, test_size: float = 0.5):
        """
        Generates dummy time series datasets for classification tasks.

        Args:
            num_samples: The number of samples to generate.
            max_ts_len: The maximum length of the time series.
            n_classes: The number of classes.
            test_size (float): The proportion of the dataset to include in the test split.

        Example:
            Easy::
                generator = TimeSeriesGenerator(num_samples=80,
                                                max_ts_len=50,
                                                n_classes=5)

                train_data, test_data = generator.generate_data()

        """
        self.num_samples = num_samples
        self.max_ts_len = max_ts_len
        self.n_classes = n_classes
        self.test_size = test_size
        self.ts_types = None

    def generate_data(self):
        """
        Generates the dataset and returns it as a tuple of train and test data.

        Returns:
            Tuple of train and test data, each containing tuples of features and targets.

        """

        ts_frame = pd.DataFrame(np.random.rand(self.num_samples, self.max_ts_len))
        labels = np.random.randint(self.n_classes, size=self.num_samples)

        X_train, X_test, y_train, y_test = train_test_split(ts_frame, labels, test_size=self.test_size, random_state=42)
        return (X_train, y_train), (X_test, y_test)


class TimeSeriesPatternGenerator:
    def __init__(self, ts_length, n_classes, n_samples):
        self.ts_length = ts_length
        self.n_classes = n_classes
        self.n_samples = n_samples

    def cycling_ts(self):
        t = np.arange(self.ts_length)
        x = np.sin(2*np.pi*t/self.ts_length)
        return x

    def random_walk_ts(self):
        x = np.cumsum(np.random.randn(self.ts_length))
        return x

    def sine_ts(self):
        t = np.arange(self.ts_length)
        x = np.sin(t)
        return x

    def square_wave_ts(self):
        t = np.arange(self.ts_length)
        x = np.where(np.sin(2*np.pi*t/self.ts_length) > 0, 1, -1)
        return x

    def sawtooth_ts(self):
        t = np.arange(self.ts_length)
        x = np.mod(t, 1)*2 - 1
        return x

    def generate_dataset(self):
        X = np.zeros((self.n_samples*self.n_classes, self.ts_length))
        y = np.zeros(self.n_samples*self.n_classes, dtype='int')
        patterns = [self.cycling_ts, self.random_walk_ts, self.sine_ts, self.square_wave_ts, self.sawtooth_ts]
        selected_patterns = np.random.choice(patterns, self.n_classes, replace=False)
        for i, pattern in enumerate(selected_patterns):
            for j in range(self.n_samples):
                X[i*self.n_samples+j,:] = pattern()
                additinal_noise = np.random.randn(self.ts_length)*0.01
                X[i*self.n_samples+j,:] += additinal_noise
                y[i*self.n_samples+j] = i
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, stratify=y)
        return (pd.DataFrame(X_train), y_train), (pd.DataFrame(X_test), y_test)


if __name__ == "__main__":

    ts_length = 100
    n_classes = 3
    n_samples = 10

    tsg = TimeSeriesPatternGenerator(ts_length, n_classes, n_samples)
    (X_train, y_train), (X_test, y_test) = tsg.generate_dataset()



from typing import Optional

import numpy as np
from fedot.core.operations.operation_parameters import OperationParameters

from fedot_ind.core.operation.transformation.basis.abstract_basis import BasisDecompositionImplementation


class FourierBasisImplementation(BasisDecompositionImplementation):
    """A class for decomposing data on the Fourier basis and evaluating the derivative of the resulting decomposition.

    Example::
        ts1 = np.random.rand(200)
        ts2 = np.random.rand(200)
        ts = [ts1, ts2]
        bss = FourierBasisImplementation({'threshold': 20000'})
        basis_multi = bss.transform(ts)
        basis_1d = bss.transform(ts1)

    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.threshold = params.get('threshold')
        self.basis = None

        self.logging_params.update({'threshold': self.threshold})

    def _decompose_signal(self, input_data):
        fourier_coef = np.fft.rfft(input_data)
        frequencies = np.fft.rfftfreq(input_data.size, d=2e-3 / input_data.size)
        fourier_coef[frequencies > self.threshold] = 0
        return np.fft.irfft(fourier_coef)

    def _transform_one_sample(self, series: np.array):
        return self._get_basis(series)

    def evaluate_derivative(self, order):
        """Evaluates the derivative of the Fourier decomposition of the given data.

        Returns:
            np.array: The derivative of the Fourier decomposition of the given data.
        """
        return np.fft.ifft(1j * np.arange(len(self.data_range)) * self.decomposed)



"""This module contains value checking functions common to the whole framework."""

from typing import Set, Any


def parameter_value_check(
        parameter: str,
        value: Any,
        valid_values: Set
) -> None:
    """Checks if the parameter value is in the set of valid values.

    Args:
        parameter: Name of the checked parameter.
        value: Value of the checked parameter.
        valid_values: Set of the valid parameter values.

    Rises:
        ValueError: If ``value`` is not in ``valid_values``.


    """
    if value not in valid_values:
        raise ValueError(
            f"{parameter} must be one of {valid_values}, but got {parameter}='{value}'"
        )



import math
from typing import Union

import numpy as np
import pandas as pd
from scipy.signal import find_peaks
from statsmodels.tsa.stattools import acf


class WindowSizeSelector:
    """Class to select appropriate window size to catch periodicity for time series analysis.

    There are two group of algorithms implemented:

    Whole-Series-Based (WSB):
        1. 'hac' - highest_autocorrelation
        2. 'dff' - dominant_fourier_frequency

    Subsequence-based (SB):
        1. 'mwf' - multi_window_finder
        2. 'sss' - summary_statistics_subsequence

    Args:
        method: by ``default``, it is 'dff'.
            You can choose between: 'hac', 'dff', 'sss' or 'mwf'.
        window_range: % of time series length, by ``default`` it is (5, 50).

    Attributes:
        length_ts(int): length of the time_series.
        window_max(int): maximum window size in real values.
        window_min(int): minimum window size in real values.
        dict_methods(dict): dictionary with all implemented methods.

    Example:
        To find window size for single time series::
            ts = np.random.rand(1000)
            ws_selector = WindowSizeSelector(method='hac')
            window_size = ws_selector.get_window_size(time_series=ts)

        To find window size for multiple time series::
            ts = np.random.rand(1000, 10)
            ws_selector = WindowSizeSelector(method='hac')
            window_size = ws_selector.apply(time_series=ts, average='median')


    Reference:
        (c) "Windows Size Selection in Unsupervised Time Series Analytics: A Review and Benchmark. Arik Ermshaus,
        Patrick Schafer, and Ulf Leser. 2022"

    """

    def __init__(self, method: str = 'dff', window_range: tuple = (5, 50)):

        assert window_range[0] < window_range[1], 'Upper bound of window range should be bigger than lower bound'

        self.dict_methods = {'hac': self.autocorrelation,
                             'dff': self.dominant_fourier_frequency,
                             'mwf': self.mwf,
                             'sss': self.summary_statistics_subsequence}
        self.wss_algorithm = method
        self.window_range = window_range
        self.window_max = None
        self.window_min = None
        self.length_ts = None

    def apply(self, time_series: Union[pd.DataFrame, np.array], average: str = 'median') -> int:
        """Method to run WSS class over selected time series in parallel mode via joblib

        Args:
            time_series: time series to study
            average: 'mean' or 'median' to average window size over all time series

        Returns:
            window_size_selected: value which has been chosen as appropriate window size
        """
        methods = {'mean': np.mean, 'median': np.median}
        assert average in methods.keys(), 'Hyperparameters error: `average` should be mean or median'

        if isinstance(time_series, pd.DataFrame):
            time_series = time_series.values

        window_list = [self.get_window_size(ts) for ts in time_series]
        return round(methods[average](window_list))

    def get_window_size(self, time_series: np.array) -> int:
        """Main function to run WSS class over selected time series

        Note:
            One of the reason of ValueError is that time series size can be equal or smaller than 50.
            In case of it try to initially set window_size min and max.

        Returns:
            window_size_selected: value which has been chosen as appropriate window size
        """
        if time_series.shape[0] == 1:  # If time series is a part of multivariate one
            time_series = np.array(time_series[0])
        self.length_ts = len(time_series)

        self.window_max = int(round(self.length_ts * self.window_range[1] / 100))  # in real values
        self.window_min = int(round(self.length_ts * self.window_range[0] / 100))  # in real values

        window_size_selected = self.dict_methods[self.wss_algorithm](time_series=time_series)
        return round(window_size_selected * 100 / self.length_ts)  # in %

    def dominant_fourier_frequency(self, time_series: np.array) -> int:
        """
        Method to find dominant fourier frequency in time series and return appropriate window size. It is based on
        the assumption that the dominant frequency is the one with the highest magnitude in the Fourier transform. The
        window size is then the inverse of the dominant frequency.

        """
        fourier = np.fft.fft(time_series)
        freq = np.fft.fftfreq(time_series.shape[0], 1)

        magnitudes, window_sizes = [], []

        for coef, freq in zip(fourier, freq):
            if coef and freq > 0:
                window_size = int(1 / freq)
                mag = math.sqrt(coef.real * coef.real + coef.imag * coef.imag)

                if self.window_min <= window_size < self.window_max:
                    window_sizes.append(window_size)
                    magnitudes.append(mag)

        return window_sizes[np.argmax(magnitudes)]

    def autocorrelation(self, time_series: np.array) -> int:
        """Method to find the highest autocorrelation in time series and return appropriate window size. It is based on
        the assumption that the lag of highest autocorrelation coefficient corresponds to the window size that best
        captures the periodicity of the time series.
        """
        ts_len = time_series.shape[0]
        acf_values = acf(time_series, fft=True, nlags=int(ts_len / 2))

        peaks, _ = find_peaks(acf_values)
        peaks = peaks[np.logical_and(peaks >= self.window_min, peaks < self.window_max)]
        corrs = acf_values[peaks]

        if peaks.shape[0] == 0:  # if there is no peaks in range (window_min, window_max) return window_min
            return self.window_min
        return peaks[np.argmax(corrs)]

    def mwf(self, time_series: np.array) -> int:
        """ Method to find the window size that minimizes the moving average residual. It is based on the assumption
        that the window size that best captures the periodicity of the time series is the one that minimizes the
        difference between the moving average and the time series.
        """

        all_averages, window_sizes = [], []

        for w in range(self.window_min, self.window_max, 1):
            movingAvg = np.array(self.movmean(time_series, w))
            all_averages.append(movingAvg)
            window_sizes.append(w)

        movingAvgResiduals = []

        for i, w in enumerate(window_sizes):
            moving_avg = all_averages[i][:len(all_averages[-1])]
            movingAvgResidual = np.log(abs(moving_avg - (moving_avg).mean()).sum())
            movingAvgResiduals.append(movingAvgResidual)

        b = (np.diff(np.sign(np.diff(movingAvgResiduals))) > 0).nonzero()[0] + 1  # local min

        if len(b) == 0:
            return self.window_min
        if len(b) < 3:
            return window_sizes[b[0]]

        reswin = np.array([window_sizes[b[i]] / (i + 1) for i in range(3)])
        w = np.mean(reswin)

        return int(w)

    def movmean(self, ts, w):
        """Fast moving average function"""
        moving_avg = np.cumsum(ts, dtype=float)
        moving_avg[w:] = moving_avg[w:] - moving_avg[:-w]
        return moving_avg[w - 1:] / w

    def summary_statistics_subsequence(self, time_series: np.array, threshold=.89) -> int:
        """Method to find the window size that maximizes the subsequence unsupervised similarity score (SUSS). It is
        based on the assumption that the window size that best captures the periodicity of the time series is the one
        that maximizes the similarity between subsequences of the time series.
        """
        # lbound = self.window_min
        time_series = (time_series - time_series.min()) / (time_series.max() - time_series.min())

        ts_mean = np.mean(time_series)
        ts_std = np.std(time_series)
        ts_min_max = np.max(time_series) - np.min(time_series)

        stats = (ts_mean, ts_std, ts_min_max)

        max_score = self.suss_score(time_series=time_series, window_size=1, stats=stats)
        min_score = self.suss_score(time_series=time_series, window_size=time_series.shape[0] - 1, stats=stats)

        exp = 0

        # exponential search (to find window size interval)
        while True:
            window_size = 2 ** exp

            if window_size < self.window_min:
                exp += 1
                continue

            score = 1 - (self.suss_score(time_series, window_size, stats) - min_score) / (max_score - min_score)

            if score > threshold:
                break

            exp += 1

        lbound, ubound = max(self.window_min, 2 ** (exp - 1)), 2 ** exp + 1

        # binary search (to find window size in interval)
        while lbound <= ubound:
            window_size = int((lbound + ubound) / 2)
            score = 1 - (self.suss_score(time_series, window_size, stats) - min_score) / (max_score - min_score)

            if score < threshold:
                lbound = window_size + 1
            elif score > threshold:
                ubound = window_size - 1
            else:
                break

        return 2 * lbound

    def suss_score(self, time_series, window_size, stats):
        roll = pd.Series(time_series).rolling(window_size)
        ts_mean, ts_std, ts_min_max = stats

        roll_mean = roll.mean().to_numpy()[window_size:]
        roll_std = roll.std(ddof=0).to_numpy()[window_size:]
        roll_min = roll.min().to_numpy()[window_size:]
        roll_max = roll.max().to_numpy()[window_size:]

        X = np.array([
            roll_mean - ts_mean,
            roll_std - ts_std,
            (roll_max - roll_min) - ts_min_max
        ])

        X = np.sqrt(np.sum(np.square(X), axis=0)) / np.sqrt(window_size)

        return np.mean(X)



import os
from pathlib import Path
from typing import Dict, List, Type, Union

import pandas as pd
from torch.utils.tensorboard import SummaryWriter


class Writer:
    """Generalized class for writing metrics.

    Args:
        path: Path for recording metrics.
    """
    def __init__(self, path: Union[str, Path]) -> None:
        self.path = path

    def write_scores(
            self,
            phase: str,
            scores: Dict[str, float],
            x,
    ) -> None:
        """Write scores from dictionary by writer.

        Args:
            phase: Experiment phase for grouping records, e.g. 'train'.
            scores: Dictionary {metric_name: value}.
            x: The independent variable.
        """
        raise NotImplementedError

    def close(self) -> None:
        """Finishing the writer."""
        pass


class TFWriter(Writer):
    """Сlass for writing metrics using SummaryWriter.

    Args:
        path: Path for recording metrics.
    """

    def __init__(self, path: Union[str, Path]) -> None:
        super().__init__(path=path)
        self.writer = SummaryWriter(path)

    def write_scores(
            self,
            phase: str,
            scores: Dict[str, float],
            x,
    ) -> None:
        """Write scores from dictionary by writer.

        Args:
            phase: Experiment phase for grouping records, e.g. 'train'.
            scores: Dictionary {metric_name: value}.
            x: The independent variable.
        """
        for key, score in scores.items():
            self.writer.add_scalar(f"{phase}/{key}", score, x)

    def close(self) -> None:
        """Finishing the writer."""
        self.writer.close()


class CSVWriter(Writer):
    """Сlass for writing metrics using Pandas .

    Args:
        path: Path for recording metrics.
    """

    def __init__(self, path: Union[str, Path]):
        super().__init__(path)
        os.makedirs(self.path, exist_ok=True)

    def write_scores(
            self,
            phase: str,
            scores: Dict[str, float],
            x,
    ) -> None:
        """Write scores from dictionary by writer.

        Args:
            phase: Experiment phase for grouping records, used as csv filename.
            scores: Dictionary {metric_name: value}.
            x: The independent variable.
        """
        data = pd.DataFrame(data=scores, index=[x])
        path = os.path.join(self.path, f'{phase}.csv')
        if os.path.exists(path):
            data.to_csv(path, mode='a', header=False)
        else:
            data.to_csv(path)


class WriterComposer(Writer):
    """Composes several writers together.

    Args:
        path: Path for recording metrics.
        writers: Types of used writers.
    """
    def __init__(self, path: Union[str, Path], writers: List[Type[Writer]]) -> None:
        super().__init__(path)
        self.writers = [writer(path=path) for writer in writers]

    def write_scores(
            self,
            phase: str,
            scores: Dict[str, float],
            x,
    ) -> None:
        """Write scores from dictionary by writer.

        Args:
            phase: Experiment phase for grouping records, used as csv filename.
            scores: Dictionary {metric_name: value}.
            x: The independent variable.
        """
        for writer in self.writers:
            writer.write_scores(phase=phase, scores=scores, x=x)

    def close(self) -> None:
        """Finishing the writer."""
        for writer in self.writers:
            writer.close()



import numpy as np
from scipy.spatial.distance import pdist, squareform


class TSTransformer:
    def __init__(self, time_series, rec_metric):
        self.time_series = time_series.reshape(-1, 1) if len(time_series.shape) == 1 else time_series
        self.recurrence_matrix = None
        self.threshold_baseline = [0.95, 0.7]  # TODO add threshold for other metrics
        self.min_signal_ratio = 0.6
        self.max_signal_ratio = 0.85
        self.rec_metric = rec_metric

    def ts_to_recurrence_matrix(self,
                                threshold=None):
        distance_matrix = pdist(metric=self.rec_metric, X=self.time_series)
        distance_matrix = np.ones(shape=distance_matrix.shape[0]) - distance_matrix
        distance_matrix = self.binarization(distance_matrix, threshold=threshold)
        self.recurrence_matrix = squareform(distance_matrix)
        return self.recurrence_matrix

    def binarization(self, distance_matrix, threshold):
        best_threshold_flag = False
        signal_ratio_list = []
        recurrence_matrix = None
        if threshold is None:
            for threshold_baseline in self.threshold_baseline:
                threshold = threshold_baseline
                tmp_array = np.copy(distance_matrix)
                tmp_array[tmp_array < threshold_baseline] = 0.0
                tmp_array[tmp_array >= threshold_baseline] = 1.0
                signal_ratio = np.where(tmp_array == 0)[0].shape[0] / tmp_array.shape[0]

                if self.min_signal_ratio < signal_ratio < self.max_signal_ratio:
                    best_ratio = signal_ratio
                    recurrence_matrix = tmp_array
                    best_threshold_flag = True
                    if signal_ratio > best_ratio:
                        recurrence_matrix = tmp_array
                else:
                    signal_ratio_list.append(abs(self.max_signal_ratio - signal_ratio))

                del tmp_array

        if not best_threshold_flag:
            distance_matrix[distance_matrix < self.threshold_baseline[0]] = 0.0
            distance_matrix[distance_matrix >= self.threshold_baseline[0]] = 1.0
            recurrence_matrix = distance_matrix
        return recurrence_matrix

    def get_recurrence_metrics(self):
        if self.recurrence_matrix is None:
            return self.ts_to_recurrence_matrix()



"""This module contains functions and classes for computing metrics
 in computer vision tasks.
 """
from abc import ABC, abstractmethod
from typing import Dict, List

import torch
from sklearn.metrics import accuracy_score, roc_auc_score
from sklearn.metrics import precision_recall_fscore_support, f1_score
from torch.nn.functional import softmax
from torchmetrics.detection.mean_ap import MeanAveragePrecision


def iou_score(
        outputs: torch.Tensor,
        masks: torch.Tensor,
        threshold: float = 0.5,
        smooth: float = 1e-10
) -> torch.Tensor:
    """Computes intersection over union (masks) on batch.

    Args:
        outputs: Output from semantic segmentation model.
        masks: True masks.
        threshold: Binarization threshold for output.
        smooth: Additional constant to avoid division by zero.

    Returns:
        Intersection over union for batch.
    """
    outputs = (outputs > threshold).float()
    intersection = torch.logical_and(outputs, masks).float().sum((2, 3))
    union = torch.logical_or(outputs, masks).float().sum((2, 3))
    iou = (intersection + smooth) / (union + smooth)
    iou[union == 0] = -1
    return iou


def dice_score(
        outputs: torch.Tensor,
        masks: torch.Tensor,
        threshold: float = 0.5,
        smooth: float = 1e-10
) -> torch.Tensor:
    """Computes dice coefficient (masks) on batch.

    Args:
        outputs: Output from semantic segmentation model.
        masks: True masks.
        threshold: Binarization threshold for output.
        smooth: Additional constant to avoid division by zero.

    Returns:
        Dice for batch.
    """
    outputs = (outputs > threshold).float()
    intersection = torch.logical_and(outputs, masks).float().sum((2, 3))
    total = (outputs + masks).sum((2, 3))
    dice = (2 * intersection + smooth) / (total + smooth)
    dice[total == 0] = -1
    return dice


class MetricCounter(ABC):
    """Generalized class for calculating metrics"""

    def __init__(self, **kwargs) -> None:
        pass

    @abstractmethod
    def update(self, **kwargs) -> None:
        """Have to implement updating, taking model outputs as input."""
        raise NotImplementedError

    @abstractmethod
    def compute(self) -> Dict[str, float]:
        """Have to implement metrics computing."""
        raise NotImplementedError


class ClassificationMetricCounter(MetricCounter):
    """Calculates metrics for classification task.

    Args:
        class_metrics:  If ``True``, calculates metrics for each class.
    """

    def __init__(self, class_metrics: bool = False) -> None:
        super().__init__()
        self.y_true = []
        self.y_pred = []
        self.y_score = []
        self.class_metrics = class_metrics

    def update(self, predictions: torch.Tensor, targets: torch.Tensor) -> None:
        """Accumulates predictions and targets."""
        self.y_true.extend(targets.tolist())
        self.y_score.extend(softmax(predictions, dim=1).tolist())
        self.y_pred.extend(predictions.argmax(1).tolist())

    def compute(self) -> Dict[str, float]:
        """Compute accuracy, precision, recall, f1, roc auc metrics.

         Returns:
              Dictionary: `{metric: score}`.
        """
        precision, recall, f1, _ = precision_recall_fscore_support(
            self.y_true, self.y_pred, average='macro'
        )
        scores = {
            'accuracy': accuracy_score(self.y_true, self.y_pred),
            'precision': precision,
            'recall': recall,
            'f1': f1,
            'roc_auc': roc_auc_score(self.y_true, self.y_score, multi_class='ovo'),
        }
        if self.class_metrics:
            f1s = f1_score(self.y_true, self.y_pred, average=None)
            scores.update({f'f1_for_class_{i}': s for i, s in enumerate(f1s)})
        return scores


class SegmentationMetricCounter(MetricCounter):
    """Calculates metrics for semantic segmentation task.

    Args:
        class_metrics:  If ``True``, calculates metrics for each class.
    """

    def __init__(self, class_metrics: bool = False) -> None:
        super().__init__()
        self.iou = []
        self.dice = []
        self.class_metrics = class_metrics

    def update(self, predictions: torch.Tensor, targets: torch.Tensor) -> None:
        """Accumulates iou and dice."""
        masks = torch.zeros_like(predictions)
        for i in range(predictions.shape[1]):
            masks[:, i, :, :] = torch.squeeze(targets == i)
        self.iou.append(iou_score(predictions, masks))
        self.dice.append(dice_score(predictions, masks))

    def compute(self) -> Dict[str, float]:
        """Compute average metrics.

         Returns:
              Dictionary: `{metric: score}`.
        """
        iou = torch.cat(self.iou).T
        dice = torch.cat(self.dice).T

        scores = {
            'iou': iou[1:][iou[1:] >= 0].mean().item(),
            'dice': dice[1:][dice[1:] >= 0].mean().item()
        }
        if self.class_metrics:
            scores.update({f'iou_for_class_{i}': s[s >= 0].mean().item() for i, s in enumerate(iou)})
            scores.update({f'dice_for_class_{i}': s[s >= 0].mean().item() for i, s in enumerate(dice)})
        return scores


class ObjectDetectionMetricCounter(MetricCounter):
    """Calculates metrics for object detection task.

    Args:
        class_metrics:  If ``True``, calculates metrics for each class.
    """

    def __init__(self, class_metrics: bool = False) -> None:
        super().__init__()
        self.map = MeanAveragePrecision(class_metrics=class_metrics)
        self.class_metrics = class_metrics

    def update(
            self,
            predictions: List[Dict[str, torch.Tensor]],
            targets: List[Dict[str, torch.Tensor]]
    ) -> None:
        """Accumulates predictions and targets."""
        self.map.update(preds=predictions, target=targets)

    def compute(self) -> Dict[str, float]:
        """Compute MAP, MAR metrics.

         Returns:
              Dictionary: `{metric: score}`.
        """

        scores = self.map.compute()
        if self.class_metrics:
            scores.update({f'map_for_class_{i}': s for i, s in enumerate(scores['map_per_class'])})
            scores.update({f'mar_100_for_class_{i}': s for i, s in enumerate(scores['mar_100_per_class'])})
        del scores['map_per_class']
        del scores['mar_100_per_class']
        return scores


class LossesAverager(MetricCounter):
    """Calculates the average loss."""

    def __init__(self) -> None:
        super().__init__()
        self.losses = None
        self.counter = 0

    def update(self, losses: Dict[str, torch.Tensor]) -> None:
        """Accumulates losses"""
        self.counter += 1
        if self.losses is None:
            self.losses = {k: v.item() for k, v in losses.items()}
        else:
            for key, value in losses.items():
                self.losses[key] += value.item()

    def compute(self) -> Dict[str, float]:
        """Compute average losses.

        Returns:
            Dictionary: `{metric: score}`.
        """
        return {k: v / self.counter for k, v in self.losses.items()}



from functools import partial

from hyperopt import hp

NESTED_PARAMS_LABEL = 'nested_label'

industrial_search_space = {
    'eigen_basis':
        {'window_size': {'hyperopt-dist': hp.choice, 'sampling-scope': [[x for x in range(5, 50, 5)]]},
         'stride': {'hyperopt-dist': hp.choice, 'sampling-scope': [[x for x in range(1, 10, 1)]]}},
    'wavelet_basis':
        {'n_components': {'hyperopt-dist': hp.uniformint, 'sampling-scope': [2, 10]},
         'wavelet': {'hyperopt-dist': hp.choice,
                     'sampling-scope': [['mexh', 'shan', 'morl', 'cmor', 'fbsp', 'db5', 'sym5']]}},
    'fourier_basis':
        {'threshold': {'hyperopt-dist': hp.uniformint, 'sampling-scope': [10000, 50000]}},
    'quantile_extractor':
        {'stride': {'hyperopt-dist': hp.choice, 'sampling-scope': [[x for x in range(1, 10, 1)]]},
         'window_size': {'hyperopt-dist': hp.choice, 'sampling-scope': [[x for x in range(0, 50, 3)]]}},
    'recurrence_extractor':
        {'window_size': (hp.uniformint, [1, 50]),
         'stride': {'hyperopt-dist': hp.choice, 'sampling-scope': [[x for x in range(1, 10, 1)]]},
         'rec_metric': (hp.choice, [['chebyshev', 'cosine', 'euclidean', 'mahalanobis']])},
    'signal_extractor':
        {'wavelet': {'hyperopt-dist': hp.choice,
                     'sampling-scope': [['mexh', 'shan', 'morl', 'cmor', 'fbsp', 'db5', 'sym5']]}},
    'ssa_forecaster':
        {'window_size_method': {'hyperopt-dist': hp.choice, 'window_size_method': [['hac', 'dff', 'mwf', 'sss']]}}
    }


def get_industrial_search_space(self):
    parameters_per_operation = {'kmeans': {
        'n_clusters': {
            'hyperopt-dist': hp.uniformint,
            'sampling-scope': [2, 7],
            'type': 'discrete'}
    },
        'adareg': {
            'learning_rate': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-3, 1],
                'type': 'continuous'},
            'loss': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [["linear", "square", "exponential"]],
                'type': 'categorical'}
        },
        'gbr': {
            'loss': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [["ls", "lad", "huber", "quantile"]],
                'type': 'categorical'},
            'learning_rate': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-3, 1],
                'type': 'continuous'},
            'max_depth': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 11],
                'type': 'discrete'},
            'min_samples_split': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 21],
                'type': 'discrete'},
            'min_samples_leaf': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 21],
                'type': 'discrete'},
            'subsample': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 1.0],
                'type': 'continuous'},
            'max_features': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 1.0],
                'type': 'continuous'},
            'alpha': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.75, 0.99],
                'type': 'continuous'}
        },
        'logit': {
            'C': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [1e-2, 10.0],
                'type': 'continuous'}
        },
        'rf': {
            'criterion': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [["gini", "entropy"]],
                'type': 'categorical'},
            'max_features': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 1.0],
                'type': 'continuous'},
            'min_samples_split': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 10],
                'type': 'discrete'},
            'min_samples_leaf': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 15],
                'type': 'discrete'},
            'bootstrap': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[True, False]],
                'type': 'categorical'}
        },
        'ridge': {
            'alpha': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.01, 10.0],
                'type': 'continuous'}
        },
        'lasso': {
            'alpha': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.01, 10.0],
                'type': 'continuous'}
        },
        'rfr': {
            'max_features': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 1.0],
                'type': 'continuous'},
            'min_samples_split': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 21],
                'type': 'discrete'},
            'min_samples_leaf': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 15],
                'type': 'discrete'},
            'bootstrap': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[True, False]],
                'type': 'categorical'}
        },
        'xgbreg': {
            'max_depth': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 11],
                'type': 'discrete'},
            'learning_rate': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-3, 1],
                'type': 'continuous'},
            'subsample': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 1.0],
                'type': 'continuous'},
            'min_child_weight': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 21],
                'type': 'discrete'},
        },
        'xgboost': {
            'max_depth': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 7],
                'type': 'discrete'},
            'learning_rate': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-3, 1],
                'type': 'continuous'},
            'subsample': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 0.99],
                'type': 'continuous'},
            'min_child_weight': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 21],
                'type': 'discrete'}
        },
        'svr': {
            'C': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [1e-4, 25.0],
                'type': 'continuous'},
            'epsilon': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [1e-4, 1],
                'type': 'continuous'},
            'tol': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-5, 1e-1],
                'type': 'continuous'},
            'loss': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [["epsilon_insensitive", "squared_epsilon_insensitive"]],
                'type': 'categorical'}
        },
        'dtreg': {
            'max_depth': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 11],
                'type': 'discrete'},
            'min_samples_split': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 21],
                'type': 'discrete'},
            'min_samples_leaf': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 21],
                'type': 'discrete'}
        },
        'treg': {
            'max_features': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 1.0],
                'type': 'continuous'},
            'min_samples_split': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 21],
                'type': 'discrete'},
            'min_samples_leaf': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 21],
                'type': 'discrete'},
            'bootstrap': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[True, False]],
                'type': 'categorical'}
        },
        'dt': {
            'max_depth': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 11],
                'type': 'discrete'},
            'min_samples_split': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 21],
                'type': 'discrete'},
            'min_samples_leaf': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 21],
                'type': 'discrete'}
        },
        'knnreg': {
            'n_neighbors': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 50],
                'type': 'discrete'},
            'weights': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [["uniform", "distance"]],
                'type': 'categorical'},
            'p': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[1, 2]],
                'type': 'categorical'}
        },
        'knn': {
            'n_neighbors': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 50],
                'type': 'discrete'},
            'weights': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [["uniform", "distance"]],
                'type': 'categorical'},
            'p': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[1, 2]],
                'type': 'categorical'}
        },
        'arima': {
            'p': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 7],
                'type': 'discrete'},
            'd': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [0, 2],
                'type': 'discrete'},
            'q': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 5],
                'type': 'discrete'}
        },
        'stl_arima': {
            'p': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 7],
                'type': 'discrete'},
            'd': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [0, 2],
                'type': 'discrete'},
            'q': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 5],
                'type': 'discrete'},
            'period': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 365],
                'type': 'discrete'}
        },
        'ar': {
            'lag_1': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [2, 200],
                'type': 'continuous'},
            'lag_2': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [2, 800],
                'type': 'continuous'}
        },
        'ets': {
            'error': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [["add", "mul"]],
                'type': 'categorical'},
            'trend': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[None, "add", "mul"]],
                'type': 'categorical'},
            'seasonal': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[None, "add", "mul"]],
                'type': 'categorical'},
            'damped_trend': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[True, False]],
                'type': 'categorical'},
            'seasonal_periods': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [1, 100],
                'type': 'continuous'}
        },
        'glm': {
            NESTED_PARAMS_LABEL: {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[
                    {
                        'family': 'gaussian',
                        'link': hp.choice('link_gaussian', ['identity',
                                                            'inverse_power',
                                                            'log'])
                    },
                    {
                        'family': 'gamma',
                        'link': hp.choice('link_gamma', ['identity',
                                                         'inverse_power',
                                                         'log'])
                    },
                    {
                        'family': 'inverse_gaussian',
                        'link': hp.choice('link_inv_gaussian', ['identity',
                                                                'inverse_power'])
                    }

                ]],
                'type': 'categorical'}
        },
        'cgru': {
            'hidden_size': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [20, 200],
                'type': 'continuous'},
            'learning_rate': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.0005, 0.005],
                'type': 'continuous'},
            'cnn1_kernel_size': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [3, 8],
                'type': 'discrete'},
            'cnn1_output_size': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[8, 16, 32, 64]],
                'type': 'categorical'},
            'cnn2_kernel_size': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [3, 8],
                'type': 'discrete'},
            'cnn2_output_size': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[8, 16, 32, 64]],
                'type': 'categorical'},
            'batch_size': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[64, 128]],
                'type': 'categorical'},
            'num_epochs': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[10, 20, 50, 100]],
                'type': 'categorical'},
            'optimizer': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [['adamw', 'sgd']],
                'type': 'categorical'},
            'loss': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [['mae', 'mse']],
                'type': 'categorical'},
        },
        'pca': {
            'n_components': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.1, 0.99],
                'type': 'continuous'}
        },
        'kernel_pca': {
            'n_components': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 20],
                'type': 'discrete'},
            'kernel': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [['linear', 'poly', 'rbf', 'sigmoid', 'cosine', 'precomputed']],
                'type': 'categorical'}
        },
        'fast_ica': {
            'n_components': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 20],
                'type': 'discrete'},
            'fun': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [['logcosh', 'exp', 'cube']],
                'type': 'categorical'}
        },
        'ransac_lin_reg': {
            'min_samples': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.1, 0.9],
                'type': 'continuous'},
            'residual_threshold': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [0.1, 1000],
                'type': 'continuous'},
            'max_trials': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [50, 500],
                'type': 'continuous'},
            'max_skips': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [50, 500000],
                'type': 'continuous'}
        },
        'ransac_non_lin_reg': {
            'min_samples': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.1, 0.9],
                'type': 'continuous'},
            'residual_threshold': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [0.1, 1000],
                'type': 'continuous'},
            'max_trials': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [50, 500],
                'type': 'continuous'},
            'max_skips': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [50, 500000],
                'type': 'continuous'}
        },
        'isolation_forest_reg': {
            'max_samples': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 0.99],
                'type': 'continuous'},
            'max_features': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 0.99],
                'type': 'continuous'},
            'bootstrap': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[True, False]],
                'type': 'categorical'}
        },
        'isolation_forest_class': {
            'max_samples': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 0.99],
                'type': 'continuous'},
            'max_features': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.05, 0.99],
                'type': 'continuous'},
            'bootstrap': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[True, False]],
                'type': 'categorical'}
        },
        'rfe_lin_reg': {
            'n_features_to_select': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.5, 0.9],
                'type': 'continuous'},
            'step': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.1, 0.2],
                'type': 'continuous'}
        },
        'rfe_non_lin_reg': {
            'n_features_to_select': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.5, 0.9],
                'type': 'continuous'},
            'step': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.1, 0.2],
                'type': 'continuous'}
        },
        'poly_features': {
            'degree': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 5],
                'type': 'discrete'},
            'interaction_only': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[True, False]],
                'type': 'categorical'}
        },
        'polyfit': {
            'degree': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 6],
                'type': 'discrete'}
        },
        'lagged': {
            'window_size': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [5, 500],
                'type': 'discrete'}
        },
        'sparse_lagged': {
            'window_size': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [5, 500],
                'type': 'discrete'},
            'n_components': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0, 0.5],
                'type': 'continuous'},
            'use_svd': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[True, False]],
                'type': 'categorical'}
        },
        'smoothing': {
            'window_size': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 20],
                'type': 'discrete'}
        },
        'gaussian_filter': {
            'sigma': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [1, 5],
                'type': 'continuous'}
        },
        'diff_filter': {
            'poly_degree': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 5],
                'type': 'discrete'},
            'order': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [1, 3],
                'type': 'continuous'},
            'window_size': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [3, 20],
                'type': 'continuous'}
        },
        'cut': {
            'cut_part': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0, 0.9],
                'type': 'continuous'}
        },
        'lgbm': {
            'class_weight': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[None, 'balanced']],
                'type': 'categorical'},
            'num_leaves': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 256],
                'type': 'discrete'},
            'learning_rate': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [0.01, 0.2],
                'type': 'continuous'},
            'colsample_bytree': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.4, 1],
                'type': 'continuous'},
            'subsample': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.4, 1],
                'type': 'continuous'},
            'reg_alpha': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-8, 10],
                'type': 'continuous'},
            'reg_lambda': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-8, 10],
                'type': 'continuous'}
        },
        'lgbmreg': {
            'num_leaves': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 256],
                'type': 'discrete'},
            'learning_rate': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [0.01, 0.2],
                'type': 'continuous'},
            'colsample_bytree': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.4, 1],
                'type': 'continuous'},
            'subsample': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.4, 1],
                'type': 'continuous'},
            'reg_alpha': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-8, 10],
                'type': 'continuous'},
            'reg_lambda': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-8, 10],
                'type': 'continuous'}
        },
        'catboost': {
            'max_depth': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 11],
                'type': 'discrete'},
            'learning_rate': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [0.01, 0.2],
                'type': 'continuous'},
            'min_data_in_leaf': {
                'hyperopt-dist': partial(hp.qloguniform, q=1),
                'sampling-scope': [0, 6],
                'type': 'discrete'},
            'border_count': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 255],
                'type': 'discrete'},
            'l2_leaf_reg': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-8, 10],
                'type': 'continuous'}
        },
        'catboostreg': {
            'max_depth': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [1, 11],
                'type': 'discrete'},
            'learning_rate': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [0.01, 0.2],
                'type': 'continuous'},
            'min_data_in_leaf': {
                'hyperopt-dist': partial(hp.qloguniform, q=1),
                'sampling-scope': [0, 6],
                'type': 'discrete'},
            'border_count': {
                'hyperopt-dist': hp.uniformint,
                'sampling-scope': [2, 255],
                'type': 'discrete'},
            'l2_leaf_reg': {
                'hyperopt-dist': hp.loguniform,
                'sampling-scope': [1e-8, 10],
                'type': 'continuous'}
        },
        'resample': {
            'balance': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [['expand_minority', 'reduce_majority']],
                'type': 'categorical'},
            'replace': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[True, False]],
                'type': 'categorical'},
            'balance_ratio': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.3, 1],
                'type': 'continuous'}
        },
        'lda': {
            'solver': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [['svd', 'lsqr', 'eigen']],
                'type': 'categorical'},
            'shrinkage': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.1, 0.9],
                'type': 'continuous'}
        },
        'ts_naive_average': {
            'part_for_averaging': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.1, 1],
                'type': 'continuous'}
        },
        'locf': {
            'part_for_repeat': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.01, 0.5],
                'type': 'continuous'}
        },
        'word2vec_pretrained': {
            'model_name': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [['glove-twitter-25', 'glove-twitter-50',
                                    'glove-wiki-gigaword-100', 'word2vec-ruscorpora-300']],
                'type': 'categorical'}
        },
        'tfidf': {
            'ngram_range': {
                'hyperopt-dist': hp.choice,
                'sampling-scope': [[(1, 1), (1, 2), (1, 3)]],
                'type': 'categorical'},
            'min_df': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.0001, 0.1],
                'type': 'continuous'},
            'max_df': {
                'hyperopt-dist': hp.uniform,
                'sampling-scope': [0.9, 0.99],
                'type': 'continuous'}
        },
    }
    for key in industrial_search_space:
        parameters_per_operation[key] = industrial_search_space[key]

    if self.custom_search_space is not None:
        for operation in self.custom_search_space.keys():
            if self.replace_default_search_space:
                parameters_per_operation[operation] = self.custom_search_space[operation]
            else:
                for key, value in self.custom_search_space[operation].items():
                    parameters_per_operation[operation][key] = value

    return parameters_per_operation



import logging
from typing import Union

import numpy as np
import pandas as pd


class DataCheck:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def _replace_inf_with_nans(input_data: np.ndarray):
        values_to_replace = [np.inf, -np.inf]
        features_with_replaced_inf = np.where(np.isin(input_data,
                                                      values_to_replace),
                                              np.nan,
                                              input_data)
        input_data = features_with_replaced_inf
        return input_data

    def _check_for_nan(self, input_data: np.ndarray) -> np.ndarray:
        """Method responsible for checking if there are any NaN values in the time series dataframe
        and replacing them with 0

        Args:
            input_data: time series dataframe with NaN values

        Returns:
            input_data: time series dataframe without NaN values

        """
        if np.any(np.isnan(input_data)):
            input_data = np.nan_to_num(input_data, nan=0)
        return input_data

    def check_data(self, input_data: pd.DataFrame, return_df: bool = False) -> Union[pd.DataFrame, np.ndarray]:
        if type(input_data) != pd.DataFrame:
            return input_data
        else:
            filled_data = input_data.apply(lambda x: self._replace_inf_with_nans(x))
            filled_data = filled_data.apply(lambda x: self._check_for_nan(x))
            if filled_data.shape[0] == input_data.shape[0] and filled_data.shape[1] == input_data.shape[1]:
                input_data = filled_data
            else:
                self.logger.error('Encontered error during extracted features checking')
                raise ValueError('Data contains NaN values')

            if return_df:
                return input_data
            else:
                return input_data.values



import numpy as np
import pandas as pd
from fedot.core.data.data import InputData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum


def init_input_data(X: pd.DataFrame, y: np.ndarray, task: str = 'classification') -> InputData:
    """Method for initialization of InputData object from pandas DataFrame and numpy array with target values.

    Args:
        X: pandas DataFrame with features
        y: numpy array with target values

    Returns:
        InputData object convenient for FEDOT framework

    """
    is_multivariate_data = True if isinstance(X.iloc[0, 0], pd.Series) else False
    if is_multivariate_data:
        input_data = InputData(idx=np.arange(len(X)),
                               features=np.array(X.values.tolist()),
                               target=y.reshape(-1, 1),
                               # task=Task(TaskTypesEnum.classification),
                               task=Task(TaskTypesEnum(task)),
                               data_type=DataTypesEnum.image)
    else:
        input_data = InputData(idx=np.arange(len(X)),
                               features=X.values,
                               target=np.ravel(y).reshape(-1, 1),
                               # task=Task(TaskTypesEnum.classification),
                               task=Task(TaskTypesEnum(task)),
                               data_type=DataTypesEnum.table)

    return input_data



# -*- coding: utf-8 -*-
from typing import Optional

import numpy as np
import pandas as pd
import tensorly as tl
from pymonad.list import ListMonad
from tensorly.decomposition import parafac

from fedot_ind.core.operation.decomposition.matrix_decomposition.power_iteration_decomposition import RSVDDecomposition
from fedot_ind.core.operation.transformation.regularization.spectrum import reconstruct_basis, \
    singular_value_hard_threshold

supported_types = (pd.Series, np.ndarray, list)


class SpectrumDecomposer:
    """Decomposes the given time series with a singular-spectrum analysis. Assumes the values of the time series are
    recorded at equal intervals.

    Args:
        data: hankel matrix of time series
        ts_length: The length of the time series
        threshold: number of components.

    """

    def __init__(self, data, ts_length, threshold: Optional[int] = None):
        if type(data) == list:
            self.rank = round(data[0].shape[0] / 10)
            self.beta = data[0].shape[0] / data[0].shape[1]
        self.ts_length = ts_length
        self.svd_estimator = RSVDDecomposition()
        self.low_rank_approximation = True
        self.thr = threshold

    def svd(self, x):
        return ListMonad(self.svd_estimator.rsvd(tensor=x,
                                                 approximation=self.low_rank_approximation,
                                                 regularized_rank=self.thr))

    def threshold(self, x):
        return ListMonad([x[0],
                          x[1][:self.thr],
                          x[2]])

    def multi_threshold(self, x):
        return ListMonad([x[1],
                          list(map(lambda x: singular_value_hard_threshold(
                              singular_values=x,
                              beta=self.beta,
                              threshold=None), x[0])),
                          x[2].T]) if self.thr is None else ListMonad([x[1][
                                                                       :,
                                                                       :self.thr],
                                                                       x[0][
                                                                       :,
                                                                       :self.thr],
                                                                       x[2][
                                                                       :,
                                                                       :self.thr].T])

    def data_driven_basis(self, x):
        return ListMonad(reconstruct_basis(x[0],
                                           x[1],
                                           x[2],
                                           ts_length=self.ts_length))

    def tensor_decomposition(self, x):
        return ListMonad(parafac(tl.tensor(x), rank=self.rank).factors)



import warnings

import numpy as np
import pandas as pd
from scipy.signal import find_peaks
from scipy.stats import entropy, linregress
from sklearn.preprocessing import MinMaxScaler

warnings.filterwarnings("ignore")


def lambda_less_zero(array: np.array) -> int:
    mask = np.array(list(map(lambda x: x < 0.01, array)), dtype=int)
    return np.sum(mask)


def q5(array: np.array) -> float:
    return np.quantile(array, 0.05)


def q25(array: np.array) -> float:
    return np.quantile(array, 0.25)


def q75(array: np.array) -> float:
    return np.quantile(array, 0.75)


def q95(array: np.array) -> float:
    return np.quantile(array, 0.95)


def diff(array: np.array) -> float:
    return np.diff(array, n=len(array) - 1)[0]


# Extra methods for quantile features extraction
def skewness(array: np.array) -> float:
    if not isinstance(array, pd.Series):
        array = pd.Series(array)

    return pd.Series.skew(array)


def kurtosis(array: np.array) -> float:
    if not isinstance(array, pd.Series):
        array = pd.Series(array)
    return pd.Series.kurtosis(array)


def n_peaks(array: np.array) -> int:
    return len(find_peaks(array)[0])


def mean_ptp_distance(array: np.array):
    peaks, _ = find_peaks(array)
    return np.mean(np.diff(peaks))


def slope(array: np.array) -> float:
    return linregress(range(len(array)), array).slope


def ben_corr(x):
    """Useful for anomaly detection applications [1][2]. Returns the correlation from first digit distribution when
     compared to the Newcomb-Benford's Law distribution [3][4].

     Args:
            x (np.array): time series to calculate the feature of

     Returns:
            float: the value of this feature


     .. math::

         P(d)=\\log_{10}\\left(1+\\frac{1}{d}\\right)

     where :math:`P(d)` is the Newcomb-Benford distribution for :math:`d` that is the leading digit of the number
     {1, 2, 3, 4, 5, 6, 7, 8, 9}.

     .. rubric:: References

     |  [1] A Statistical Derivation of the Significant-Digit Law, Theodore P. Hill, Statistical Science, 1995
     |  [2] The significant-digit phenomenon, Theodore P. Hill, The American Mathematical Monthly, 1995
     |  [3] The law of anomalous numbers, Frank Benford, Proceedings of the American philosophical society, 1938
     |  [4] Note on the frequency of use of the different digits in natural numbers, Simon Newcomb, American Journal of
     |  mathematics, 1881

    """
    x = np.asarray(x)

    # retrieve first digit from data
    x = np.array(
        [int(str(np.format_float_scientific(i))[:1]) for i in np.abs(np.nan_to_num(x))]
    )

    # benford distribution
    benford_distribution = np.array([np.log10(1 + 1 / n) for n in range(1, 10)])

    data_distribution = np.array([(x == n).mean() for n in range(1, 10)])

    # np.corrcoef outputs the normalized covariance (correlation) between benford_distribution and data_distribution.
    # In this case returns a 2x2 matrix, the  [0, 1] and [1, 1] are the values between the two arrays
    return np.corrcoef(benford_distribution, data_distribution)[0, 1]


def interquartile_range(array: np.array) -> float:
    return q75(array) - q25(array)


def energy(array: np.array) -> float:
    return np.sum(np.power(array, 2)) / len(array)


def autocorrelation(array: np.array) -> float:
    """Autocorrelation of the time series with its lagged version
    """
    lagged_ts = np.roll(array, 1)
    return np.corrcoef(array, lagged_ts)[0, 1]


def zero_crossing_rate(array: np.array) -> float:
    """Returns the rate of sign-changes of the time series for a scaled version of it.
    """
    scaler = MinMaxScaler(feature_range=(-1, 1))
    scaled_array = scaler.fit_transform(array.reshape(-1, 1)).flatten()
    signs = np.sign(scaled_array)
    signs[signs == 0] = -1
    return np.sum((signs[1:] - signs[:-1]) != 0) / len(scaled_array)


def shannon_entropy(array: np.array) -> float:
    """Returns the Shannon Entropy of the time series.
    """
    p = np.unique(array, return_counts=True)[1] / len(array)
    return -np.sum(p * np.log2(p))


def base_entropy(array: np.array) -> float:
    """Returns the Shannon Entropy of the time series.
    """
    # Normalize the time series to sum up to 1
    normalized_series = array / np.sum(array)
    return entropy(normalized_series)

def ptp_amp(array: np.array) -> float:
    """Returns the peak-to-peak amplitude of the time series.
    """
    return np.ptp(array)


def crest_factor(array: np.array) -> float:
    """Returns the crest factor of the time series.
    """
    return np.max(np.abs(array)) / np.sqrt(np.mean(np.square(array)))


def mean_ema(array: np.array) -> float:
    """Returns the exponential moving average of the time series.
    """
    span = int(len(array) / 10)
    if span in [0, 1]:
        span = 2
    return pd.Series(array).ewm(span=span).mean().iloc[-1]


def mean_moving_median(array: np.array) -> float:
    span = int(len(array) / 10)
    if span in [0, 1]:
        span = 2
    return pd.Series(array).rolling(window=span, center=False).median().mean()


def hjorth_mobility(array):
    # Compute the first-order differential sequence
    diff_sequence = np.diff(array)
    # Calculate the mean power of the first-order differential sequence
    M2 = np.sum(np.power(diff_sequence, 2)) / len(diff_sequence)
    # Calculate the total power of the time series
    TP = np.sum(np.power(array, 2)) / len(array)
    # Calculate Hjorth mobility
    mobility = np.sqrt(M2 / TP)
    return mobility


def hjorth_complexity(array):
    # Compute the first-order differential sequence
    diff_sequence = np.diff(array)
    # Calculate the mean power of the first-order differential sequence
    M2 = np.sum(np.power(diff_sequence, 2)) / len(diff_sequence)
    # Calculate the total power of the time series
    TP = np.sum(np.power(array, 2)) / len(array)
    # Calculate the fourth central moment of the first-order differential sequence
    M4 = sum([(diff_sequence[i] - diff_sequence[i - 1]) ** 2 for i in range(1, len(diff_sequence))]) / len(
        diff_sequence)
    # Calculate Hjorth complexity
    complexity = np.sqrt((M4 * TP) / (M2 * M2))
    # complexity = (M4 * TP) / (M2 * M2)
    return complexity


def hurst_exponent(array):
    """ Compute the Hurst Exponent of the time series. The Hurst exponent is used as a measure of long-term memory of
    time series. It relates to the autocorrelations of the time series, and the rate at which these decrease as the
    lag between pairs of values increases. For a stationary time series, the Hurst exponent is equivalent to the
    autocorrelation exponent.

    Args:
        array: Time series

    Returns:
        hurst_exponent: Hurst exponent of the time series

    Notes:
        Author of this function is Xin Liu

    """
    X = np.array(array)
    N = X.size
    T = np.arange(1, N + 1)
    Y = np.cumsum(X)
    Ave_T = Y / T

    S_T = np.zeros(N)
    R_T = np.zeros(N)

    for i in range(N):
        S_T[i] = np.std(X[:i + 1])
        X_T = Y - T * Ave_T[i]
        R_T[i] = np.ptp(X_T[:i + 1])

    R_S = R_T / S_T
    R_S = np.log(R_S)[1:]
    n = np.log(T)[1:]
    A = np.column_stack((n, np.ones(n.size)))
    [m, c] = np.linalg.lstsq(A, R_S,rcond=None)[0]
    H = m
    return H


def pfd(X, D=None):
    """The Petrosian fractal dimension (PFD) is a chaotic algorithm used to calculate EEG signal complexity
    Compute Petrosian Fractal Dimension of a time series from either two
    cases below:
        1. X, the time series of type list (default)
        2. D, the first order differential sequence of X (if D is provided,
           recommended to speed up)

    In case 1, D is computed using Numpy's difference function.

    To speed up, it is recommended to compute D before calling this function
    because D may also be used by other functions whereas computing it here
    again will slow down.
    """
    if D is None:
        D = np.diff(X)
        D = D.tolist()
    N_delta = 0  # number of sign changes in derivative of the signal
    for i in range(1, len(D)):
        if D[i] * D[i - 1] < 0:
            N_delta += 1
    n = len(X)
    return np.log10(n) / (
        np.log10(n) + np.log10(n / n + 0.4 * N_delta)
    )
